# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django import forms
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
import django_pettycash.models as dpcmodels
import django_pettycash.utils as dpcutils


class _FormControlCSSClass(object):
    def __init__(self, *args, **kwargs):
        super(_FormControlCSSClass, self).__init__(*args, **kwargs)
        for field in self.fields:
            if isinstance(self.fields[field].widget, (forms.Select, forms.Textarea, forms.TextInput, forms.NumberInput)):
                if 'class' in self.fields[field].widget.attrs:
                    self.fields[field].widget.attrs['class'] += ' form-control'
                else:
                    self.fields[field].widget.attrs['class'] = 'form-control'


class _ForceISOFormatForDates(object):
    """
    Force the date format of DateFields in the form to YYYY-MM-DD.
    """
    def __init__(self, *args, **kwargs):
        super(_ForceISOFormatForDates, self).__init__(*args, **kwargs)
        for field in self.fields:
            if isinstance(self.fields[field], forms.DateField):
                self.fields[field].input_formats = ["%Y-%m-%d"]
                self.fields[field].widget.format = "%Y-%m-%d"


# refs #5072
class _ForceFormatForDates(object):
    """
    Force the date format of DateFields in the form to settings.DATEFORMAT,
    defaulting to YYYY-MM-DD.
    """
    def __init__(self, *args, **kwargs):
        super(_ForceFormatForDates, self).__init__(*args, **kwargs)
        for field in self.fields:
            if isinstance(self.fields[field], forms.DateField):
                date_format = getattr(settings, 'DATEFORMAT', '%Y-%m-%d')
                self.fields[field].input_formats = [date_format]
                self.fields[field].widget.format = date_format


class EntryModelForm(_FormControlCSSClass, _ForceFormatForDates, forms.ModelForm):
    class Meta:
        model = dpcmodels.Entry
        exclude = ('created_by', 'modified_by')
        fields = (
            'company', 'source', 'kind', 'amount',
            'reason', 'note', 'date', 'tags',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['source'].choices = [(None, "-------")] + [
            (x.pk, str(x)) for x in dpcmodels.Source.objects\
            .filter(monthly_balances__isnull=False)\
            .distinct()
        ]
        self.fields['amount'].widget.attrs.update({
            'min': "0.01",
        })

    def clean_source(self):
        source = self.cleaned_data.get('source')
        if source.monthly_balances.all().count() == 0:
            raise forms.ValidationError(_("This Source does not have any MonthlyBalance object related to it."))
        return source

    # refs #5523
    # refs #5814
    def create_monthly_balance_objects_if_needed(self):
        # Check if there is one MonthlyBalance for the current month
        (first_day_current_month, last_day_current_month) = dpcutils.monthrange(self.cleaned_data['date'])
        # source = self.cleaned_data['source']
        # refs #5939: Do it for every source, to ensure that when visiting DPCEntryListView all Source
        # objects have the related MonthlyBalance, and avoid template rendering problems
        for source in dpcmodels.Source.objects.all():
            try:
                dpcmodels.MonthlyBalance.objects.get(
                    date=first_day_current_month,
                    source=source,
                )
            except dpcmodels.MonthlyBalance.DoesNotExist:
                earliest_monthly_balance = None
                latest_monthly_balance = None
                # Get the latest MonthlyBalance in the past wrt date
                monthly_balances = dpcmodels.MonthlyBalance.objects\
                    .filter(
                        date__lt=first_day_current_month,
                        source=source,
                    )
                if monthly_balances.exists():
                    latest_monthly_balance = monthly_balances.latest('date')
                if latest_monthly_balance:
                    previous_end = latest_monthly_balance.end
                    d = (latest_monthly_balance.date + datetime.timedelta(days=35)).replace(day=1)
                    while d <= first_day_current_month:
                        dpcmodels.MonthlyBalance.objects.create(
                            source=source,
                            date=d,
                            start=previous_end,
                            end=previous_end,
                        )
                        d = (d + datetime.timedelta(days=35)).replace(day=1)
                # There is no previous MonthlyBalance, so search for future ones
                else:
                    monthly_balances = dpcmodels.MonthlyBalance.objects\
                        .filter(
                            date__gt=first_day_current_month,
                            source=source,
                        )
                    if monthly_balances.exists():
                        earliest_monthly_balance = monthly_balances.earliest('date')
                    if earliest_monthly_balance:
                        next_start = earliest_monthly_balance.start
                        d = (earliest_monthly_balance.date - datetime.timedelta(days=1)).replace(day=1)
                        while d >= first_day_current_month:
                            dpcmodels.MonthlyBalance.objects.create(
                                source=source,
                                date=d,
                                start=next_start,
                                end=next_start,
                            )
                            d = (d - datetime.timedelta(days=1)).replace(day=1)


class SourceModelForm(_FormControlCSSClass, _ForceISOFormatForDates, forms.ModelForm):
    fields = ['name', 'description']
    class Meta:
        model = dpcmodels.Source
        fields = ["name", "description"]


# refs #11813
class SourceCreateModelForm(_FormControlCSSClass, _ForceISOFormatForDates, forms.ModelForm):
    start_balance = forms.DecimalField(max_digits=18, decimal_places=2, label=_("Beginning balance"))

    class Meta:
        model = dpcmodels.Source
        fields = ["name", "description", "start_balance"]


class TagModelForm(_FormControlCSSClass, _ForceISOFormatForDates, forms.ModelForm):
    fields = ['name']
    class Meta:
        model = dpcmodels.Tag
        fields = ["name"]


class MonthlyBalanceModelForm(_FormControlCSSClass, _ForceISOFormatForDates, forms.ModelForm):
    class Meta:
        model = dpcmodels.MonthlyBalance
        fields = ["source", "date", "start", "end"]


class CompanyModelForm(_FormControlCSSClass, forms.ModelForm):
    class Meta:
        model = dpcmodels.get_company_model()
        exclude = []


# refs #4921
class StatsBalanceByTagForm(_FormControlCSSClass, _ForceFormatForDates, forms.Form):
    tags = forms.ModelMultipleChoiceField(queryset=dpcmodels.Tag.objects.all())
    from_date = forms.DateField(required=False)
    to_date = forms.DateField(required=False)

    def clean(self):
        cleaned_data = super(StatsBalanceByTagForm, self).clean()
        from_date = cleaned_data.get('from_date', None)
        to_date = cleaned_data.get('to_date', None)
        # refs #5076: If the dates are in the wrong order, swap them
        if from_date and to_date and from_date > to_date:
            cleaned_data['from_date'] = self.fields['from_date'].initial = to_date
            cleaned_data['to_date'] = self.fields['to_date'].initial = from_date
        return cleaned_data
