# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import datetime
from django.conf import settings
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _
from . import consts as dpc_consts



COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')
# refs #4775
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', "auth.User")


# refs #9940
def get_company_model():
    from django.apps import apps
    company_app, company_model = COMPANY_MODEL.split(".")
    app_config = apps.get_app_config(company_app)
    return app_config.get_model(model_name=company_model)


class Source(models.Model):
    """
    A source of entries in the petty cash book (like 'bank account' or 'safe')
    """
    name = models.CharField(max_length=64, verbose_name=_("Name of the source"))
    description = models.TextField(verbose_name=_("Description"))

    class Meta:
        verbose_name_plural = _("sources")

    def __str__(self):
        return self.name

    # refs #11832 #11885
    def get_last_monthly_balance(self):
        return self.monthly_balances.order_by("date").last()


class Tag(models.Model):
    """
    A tag added to an entry
    """
    name = models.CharField(max_length=32, verbose_name=_("Name of the tag"))

    class Meta:
        verbose_name_plural = _("tags")

    def __str__(self):
        return self.name


class Entry(models.Model):
    """
    An entry in the petty cash book
    """
    # refs #4811: Entry.company is not mandatory
    # The South migration had the backwards migration disabled, as this was the first migration I think it's the better choice,
    # the null=False -> True cannot be rolled back, but it's the first migration, so no problem.
    company = models.ForeignKey(COMPANY_MODEL, null=True, blank=True, related_name="entries", verbose_name=_("Company"))
    # refs #5814
    source = models.ForeignKey("Source", related_name="entries", verbose_name=_("Source of the entry"))
    kind = models.IntegerField(choices=dpc_consts.ENTRY_KIND_CHOICES, verbose_name=_("Type of the movement"))
    # Does not seem that the validators are enforcing the limits at the model level...
    amount = models.DecimalField(max_digits=18, decimal_places=2, validators=[MinValueValidator(0.01)], verbose_name=_("Amount"))
    reason = models.CharField(max_length=128, verbose_name=_("Reason"))
    date = models.DateField(verbose_name=_("Date"))
    tags = models.ManyToManyField("Tag", related_name="entries", blank=True, verbose_name=_("Tags"))
    # Automatic timestamps
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Date and time of creation"))
    modified = models.DateTimeField(auto_now=True, verbose_name=_("Date and time of last modification"))
    # refs #4775
    created_by = models.ForeignKey(AUTH_USER_MODEL, related_name='created_entries', null=True, blank=True, verbose_name=_("User that created the entry"))
    modified_by = models.ForeignKey(AUTH_USER_MODEL, related_name='modified_entries', null=True, blank=True, verbose_name=_("User that last modified the entry"))
    # refs #4880
    note = models.CharField(max_length=128, blank=True, verbose_name=_("Note"))

    class Meta:
        verbose_name_plural = _("entries")
        # refs #5065
        ordering = ('date', 'pk')

    def __str__(self):
        sign = "+" if self.kind == dpc_consts.ENTRY_KIND_INBOUND else "-"
        s = "%s %s%s € - %s" % (self.source, sign, self.amount, self.date)
        return s

    # refs #11704
    def rollback_entry(self):
        day = datetime.date(self.date.year, self.date.month, 1)
        monthly_balances = MonthlyBalance.objects.filter(
            source=self.source_before_save,
            date__gte=day,
        ).order_by('date')
        # Rollback
        if monthly_balances:
            monthly_balance = monthly_balances.first()
            if self.kind_before_save == dpc_consts.ENTRY_KIND_INBOUND:
                # The entry was INBOUND, so to rollback it, make it OUTBOUND
                monthly_balance.end -= self.amount_before_save
            else:
                # The entry was OUTBOUND, so to rollback it, make it INBOUND
                monthly_balance.end += self.amount_before_save
            # Save this monthly balance
            monthly_balance.save()
            # And then update all subsequent monthly balances
            for mb in monthly_balances[1:]:
                mb.update_balance()


# https://stackoverflow.com/questions/11843747/get-previous-value-of-a-field-in-post-save-signal
def save_fields_before_save(sender, instance, *args, **kwargs):
    try:
        instance.date_before_save = sender.objects.get(pk=instance.pk).date
        # refs #11704
        instance.source_before_save = sender.objects.get(pk=instance.pk).source
        instance.kind_before_save = sender.objects.get(pk=instance.pk).kind
        instance.amount_before_save = sender.objects.get(pk=instance.pk).amount
    except sender.DoesNotExist:
        return

def fix_monthly_balance_objects_when_saving_entry(sender, instance, created, **kwargs):
    day = datetime.date(
        day=1,
        month=instance.date.month,
        year=instance.date.year,
    )
    if not created:
        day = min([day, instance.date_before_save.replace(day=1)])
        # Check if the Entry has to be rolled back. This is to be done only when the Entry
        # is modified and the Entry.source field changed, as the other cases (change of kind
        # or amount) are taken care of in the MonthlyBalance.update_balance() method.
        if instance.source != instance.source_before_save:
            instance.rollback_entry()
    # Get all the Monthly Balance objects from this month on,
    # sorted by date, so to update them one after the other
    monthly_balances = MonthlyBalance.objects.filter(
        source=instance.source,
        date__gte=day.replace(day=1),
    ).order_by('date')
    # Update the balances
    for monthly_balance in monthly_balances:
        monthly_balance.update_balance()

models.signals.pre_save.connect(save_fields_before_save, sender=Entry)
models.signals.post_save.connect(fix_monthly_balance_objects_when_saving_entry, sender=Entry)


# refs #11704
def fix_monthly_balance_objects_when_deleting_entry(sender, instance, **kwargs):
    day = instance.date.replace(day=1)
    # Rollback the instance
    # The Entry is already deleted from the database, so the only thing needed here is the
    # update_balance() of all the MonthlyBalance objects, without the need of the pre_delete and
    # all the things that were needed in the pre_save/post_save steps.
    monthly_balances = MonthlyBalance.objects.filter(
        source=instance.source,
        date__gte=day,
    ).order_by('date')
    for mb in monthly_balances:
        mb.update_balance()

models.signals.post_delete.connect(fix_monthly_balance_objects_when_deleting_entry, sender=Entry)


class MonthlyBalance(models.Model):
    """
    A monthly balance for a source, with the date of the first day of the month, and the balance at the start
    and at the end of the month.
    """
    source = models.ForeignKey("Source", related_name="monthly_balances", verbose_name=_("Source"))
    date = models.DateField(verbose_name=_("First day of the month"))
    # The member names are not supposed to change, but the verbose_name strings can, obviously
    start = models.DecimalField(max_digits=18, decimal_places=2, verbose_name=_("Beginning balance"))
    end = models.DecimalField(max_digits=18, decimal_places=2, verbose_name=_("Final balance"), null=True, blank=True)

    class Meta:
        verbose_name_plural = _("monthly balances")

    def __str__(self):
        return "%s - %s - [%s € - %s €]" % (self.source, self.date, self.start, self.end)

    # refs #11832
    def is_first(self):
        return self.source.monthly_balances.order_by("date").first() == self

    def month(self):
        return "%s" % self.date.strftime("%B %Y")

    def calculate_previous_end(self):
        from . import utils as dpcutils
        previous_month = dpcutils.get_previous_month(self.date)
        try:
            # refs #5821
            previous_balance = MonthlyBalance.objects\
                .filter(
                    source=self.source,
                    date__lte=previous_month)\
                .latest('date')
        except:
            return None
        if previous_balance.end is None:
            previous_balance.update_balance(start=previous_balance.start)
        return previous_balance.end

    def update_balance(self, start=None):
        # Update start, if different
        previous_end = self.calculate_previous_end()
        if previous_end is not None and self.start != previous_end:
            self.start = previous_end
        if (start and self.start) and (start != self.start):
            # TODO: Do the exception creation in a better way
            raise BaseException(
                _("MonthlyBalance.start is %{before}f and you wanted to change it to %{after}f")\
                .format({'before': self.start, 'after': start})
            )
        from . import utils as dpcutils
        (first, last) = dpcutils.monthrange(self.date)
        entries = self.source.entries.filter(date__range=(first, last)).select_related()
        inbound = entries.filter(kind=dpc_consts.ENTRY_KIND_INBOUND).aggregate(models.Sum('amount'))['amount__sum'] or 0
        outbound = entries.filter(kind=dpc_consts.ENTRY_KIND_OUTBOUND).aggregate(models.Sum('amount'))['amount__sum'] or 0
        self.end = self.start + inbound - outbound
        self.save()

def fix_monthly_balance_date_before_save(sender, instance, *args, **kwargs):
    try:
        instance.date = instance.date.replace(day=1)
    except sender.DoesNotExist:
        return

def fix_monthly_balance_objects_when_saving_monthly_balance(sender, instance, created, **kwargs):
    if created:
        # Get all the subsequent MonthlyBalance objects
        source = instance.source
        for mb in source.monthly_balances.filter(date__gt=instance.date.replace(day=1)).order_by("date"):
            mb.update_balance()

models.signals.pre_save.connect(fix_monthly_balance_date_before_save, sender=MonthlyBalance)
models.signals.post_save.connect(fix_monthly_balance_objects_when_saving_monthly_balance, sender=MonthlyBalance)
