# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase
from django.test.client import Client
from model_mommy import mommy
import django_pettycash.models as dpc_models


# This is a string in the form 'app_name.model_name'
COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')
Company = apps.get_model(
    COMPANY_MODEL.split('.')[0],
    COMPANY_MODEL.split('.')[1],
)

DATEFORMAT = getattr(settings, 'DATEFORMAT')

# Import after creation of 'Company' (the Company model) otherwise it does not work
import django_pettycash.forms as dpc_forms


class EntryFormTests(TestCase):
    """
    Class to test EntryForm
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        # We need a Source
        self.source = dpc_models.Source.objects.create(name="test_source")
        # We need a Company
        # get_model is deprecated in Django 1.7, but
        # https://docs.djangoproject.com/en/1.7/ref/applications/#django.apps.AppConfig.get_model
        # does not work in Django 1.6
        self.company = mommy.make(Company)
        self.today = datetime.date.today()
        self.day = datetime.date(
            day=1,
            month=self.today.month,
            year=self.today.year,
        )

    # refs #5814
    def test_entrymodelform_source_field_choices_contains_only_sources_with_at_least_one_monthlybalance(self):
        source_without_monthlybalance = dpc_models.Source.objects.create(name="Source without")
        source_with_monthlybalance = dpc_models.Source.objects.create(name="Source with")
        dpc_models.MonthlyBalance.objects.create(
            source=source_with_monthlybalance,
            date=self.day,
            start=100,
        )
        form = dpc_forms.EntryModelForm()
        choices = [x[0] for x in form.fields['source'].choices if x[0]]
        self.assertIn(source_with_monthlybalance.pk, choices)
        self.assertNotIn(source_without_monthlybalance.pk, choices)

    def test_entrymodelform_source_field_choices_contains_only_sources_with_at_least_one_monthlybalance_but_distinct(self):
        source_without_monthlybalance = dpc_models.Source.objects.create(name="Source without")
        source_with_monthlybalance = dpc_models.Source.objects.create(name="Source with")
        dpc_models.MonthlyBalance.objects.create(
            source=source_with_monthlybalance,
            date=self.day,
            start=100,
        )
        dpc_models.MonthlyBalance.objects.create(
            source=source_with_monthlybalance,
            date=(self.day - datetime.timedelta(days=1)).replace(day=1),
            start=100,
        )
        form = dpc_forms.EntryModelForm()
        choices = [x[0] for x in form.fields['source'].choices if x[0]]
        self.assertEqual(len(choices), 1)
        self.assertIn(source_with_monthlybalance.pk, choices)
        self.assertNotIn(source_without_monthlybalance.pk, choices)
