# coding: utf-8
# Copyright (c) 2015, 2018, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Permission, User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
from unittest import skip
from model_mommy import mommy
import django_pettycash.consts as dpc_consts
import django_pettycash.models as dpc_models
import django_pettycash.utils as dpc_utils


# This is a string in the form 'app_name.model_name'
COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')

DATEFORMAT = getattr(settings, 'DATEFORMAT')

class EntryTests(TestCase):
    """
    Class to test validations in Entry model
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        # We need a Source
        self.source = dpc_models.Source.objects.create(name="test_source")
        # We need a Company
        self.company_model = apps.get_model(
            app_label=COMPANY_MODEL.split('.')[0],
            model_name=COMPANY_MODEL.split('.')[1],
        )
        self.company = mommy.make(
            self.company_model,
            name="test_company",
        )
        self.today = datetime.date.today()
        self.day = self.today.replace(day=1)
        self.month_minus_one = (self.day - datetime.timedelta(days=1)).replace(day=1)
        self.month_minus_two = (self.month_minus_one - datetime.timedelta(days=1)).replace(day=1)

    def test_entry_post_save_hook_updates_monthly_balance_with_view(self):
        mb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_pk = mb.pk
        # Test that the hook works when using the view
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': self.today.strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        mb = dpc_models.MonthlyBalance.objects.get(pk=mb_pk)
        self.assertEqual(mb.end, 110.00)

    def test_entry_post_save_hook_updates_monthly_balance_without_view(self):
        mb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_pk = mb.pk
        # Test that the hook works when NOT using the view
        dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_OUTBOUND,
            reason='test reason',
            amount=10.00,
        )
        mb = dpc_models.MonthlyBalance.objects.get(pk=mb_pk)
        self.assertEqual(mb.end, 90.00)

    def test_entry_post_save_hook_updates_the_right_monthly_balance_with_view(self):
        # Current month MonthlyBalance
        cmb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        cmb_pk = cmb.pk
        # Previous month MonthlyBalance
        pmb = dpc_models.MonthlyBalance.objects.create(
            date=dpc_utils.get_previous_month(self.day),
            source=self.source,
            start=10.00,
            end=100.00,
        )
        pmb_pk = pmb.pk
        # Test that the hook works when using the view
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': self.today.strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        cmb = dpc_models.MonthlyBalance.objects.get(pk=cmb_pk)
        self.assertEqual(cmb.end, 110.00)
        pmb = dpc_models.MonthlyBalance.objects.get(pk=pmb_pk)
        self.assertEqual(pmb.end, 100.00)

    def test_entry_post_save_hook_updates_the_right_monthly_balance_without_view(self):
        # Current month MonthlyBalance
        cmb = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        cmb_pk = cmb.pk
        # Previous month MonthlyBalance
        pmb = dpc_models.MonthlyBalance.objects.create(
            date=dpc_utils.get_previous_month(self.day),
            source=self.source,
            start=10.00,
            end=100.00,
        )
        pmb_pk = pmb.pk
        # Test that the hook works when NOT using the view
        dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_OUTBOUND,
            reason='test reason',
            amount=10.00,
        )
        cmb = dpc_models.MonthlyBalance.objects.get(pk=cmb_pk)
        self.assertEqual(cmb.end, 90.00)
        pmb = dpc_models.MonthlyBalance.objects.get(pk=pmb_pk)
        self.assertEqual(pmb.end, 100.00)

    # refs #5821
    def test_entry_post_save_hook_updates_all_future_monthly_balances_without_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month4 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month4.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3.delete()
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        self.assertEqual(entry.amount, 10)
        self.assertEqual(entry.kind, dpc_consts.ENTRY_KIND_INBOUND)
        # (Re)load objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month4 = dpc_models.MonthlyBalance.objects.get(pk=mb_month4.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month4.start, 110)
        self.assertEqual(mb_month4.end, 110)

    # refs #5821
    def test_entry_post_save_hook_updates_all_future_monthly_balances_with_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month4 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month4.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month3.delete()
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'date': mb_month1.date.replace(day=7).strftime(DATEFORMAT),
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'reason': 'test reason',
            'amount': 10.00,
        })
        self.assertEqual(response.status_code, 302)
        # (Re)load objects
        entry = dpc_models.Entry.objects.get(date=mb_month1.date.replace(day=7))
        self.assertEqual(entry.amount, 10)
        self.assertEqual(entry.kind, dpc_consts.ENTRY_KIND_INBOUND)
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month4 = dpc_models.MonthlyBalance.objects.get(pk=mb_month4.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month4.start, 110)
        self.assertEqual(mb_month4.end, 110)

    # refs #5824
    def test_entry_post_save_hook_updates_from_the_right_date_in_case_of_modification_without_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)
        # The real test starts here
        # Edit the entry
        entry.date = mb_month2.date.replace(day=7)
        entry.save()
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Recheck the balances
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 100)
        self.assertEqual(mb_month2.start, 100)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)

    # refs #5824
    def test_entry_post_save_hook_updates_from_the_right_date_in_case_of_modification_with_view(self):
        dpc_models.MonthlyBalance.objects.all().delete()
        mb_month3 = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month2 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month3.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        mb_month1 = dpc_models.MonthlyBalance.objects.create(
            date=(mb_month2.date - datetime.timedelta(days=1)).replace(day=1),
            source=self.source,
            start=100.00,
            end=100.00,
        )
        self.assertEqual(dpc_models.MonthlyBalance.objects.all().count(), 3)
        entry = dpc_models.Entry.objects.create(
            company=self.company,
            source=self.source,
            date=mb_month1.date.replace(day=7),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            reason='test reason',
            amount=10.00,
        )
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Check objects
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 110)
        self.assertEqual(mb_month2.start, 110)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)
        # The real test starts here
        # Edit the entry
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename="change_entry"))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': entry.pk}), {
            'company': entry.company.pk,
            'source': entry.source.pk,
            'date': mb_month2.date.replace(day=7).strftime(DATEFORMAT),
            'kind': entry.kind,
            'reason': entry.reason,
            'amount': entry.amount,
        })
        self.assertEqual(response.status_code, 302)
        # Reload objects
        mb_month1 = dpc_models.MonthlyBalance.objects.get(pk=mb_month1.pk)
        mb_month2 = dpc_models.MonthlyBalance.objects.get(pk=mb_month2.pk)
        mb_month3 = dpc_models.MonthlyBalance.objects.get(pk=mb_month3.pk)
        # Recheck the balances
        self.assertEqual(mb_month1.start, 100)
        self.assertEqual(mb_month1.end, 100)
        self.assertEqual(mb_month2.start, 100)
        self.assertEqual(mb_month2.end, 110)
        self.assertEqual(mb_month3.start, 110)
        self.assertEqual(mb_month3.end, 110)

    @skip("This test does not work because the amount field it's not supposed to allow negative values")
    def test_force_positive_amount(self):
        dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        e = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=100,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        e.save()
        self.assertTrue(e.amount > 0)
        e = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=-100,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        e.save()
        self.assertTrue(e.amount < 0)
        dpc_models.Entry.objects.all().delete()

    # refs #11704
    def test_monthly_balance_update_when_changing_source_of_entry(self):
        """
        Source with 100.00€ monthly balance
        Source with 101.00€ monthly balance
        Entry of 50.00€ INBOUND in the first source
        Check that the first source is 150.00€ and the second is 101.00€
        Move the entry to the other source
        Check that the first source is back to 100.00€ and the second is now 151.00€
        """
        # We need another Source
        another_source = dpc_models.Source.objects.create(name="another_test_source")
        # And MonthlyBalance objects
        monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=another_source,
            start=101.00,
            end=101.00,
        )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances
        self.assertEqual(monthly_balance.end, 100 + 50)
        self.assertEqual(another_source_monthly_balance.end, 101)
        # Move the Entry to the other source
        entry.source = another_source
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances *after* the Entry modification
        self.assertEqual(monthly_balance.end, 100)
        self.assertEqual(another_source_monthly_balance.end, 101 + 50)

    # refs #11704
    def test_monthly_balance_update_when_changing_amount_of_entry(self):
        """
        Source with 100.00€ monthly balance
        Source with 101.00€ monthly balance
        Entry of 50.00€ INBOUND in the first source
        Check that the first source is 150.00€ and the second is 101.00€
        Change the amount of the entry to 64.00€
        Check that the first source is 164.00€ and the second is 101.00€
        """
        # We need another Source
        another_source = dpc_models.Source.objects.create(name="another_test_source")
        # And MonthlyBalance objects
        monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=another_source,
            start=101.00,
            end=101.00,
        )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances
        self.assertEqual(monthly_balance.end, 100 + 50)
        self.assertEqual(another_source_monthly_balance.end, 101)
        # Change the amount of the Entry
        entry.amount = 64
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances *after* the Entry modification
        self.assertEqual(monthly_balance.end, 100 + 64)
        self.assertEqual(another_source_monthly_balance.end, 101)

    # refs #11704
    def test_monthly_balance_update_when_changing_kind_of_entry(self):
        """
        Source with 100.00€ monthly balance
        Source with 101.00€ monthly balance
        Entry of 50.00€ INBOUND in the first source
        Check that the first source is 150.00€ and the second is 101.00€
        Make the Entry OUTBOUND
        Check that the first source is 50.00€ and the second is 101.00€
        """
        # We need another Source
        another_source = dpc_models.Source.objects.create(name="another_test_source")
        # And MonthlyBalance objects
        monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=another_source,
            start=101.00,
            end=101.00,
        )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances
        self.assertEqual(monthly_balance.end, 100 + 50)
        self.assertEqual(another_source_monthly_balance.end, 101)
        # Change the kind of the Entry
        entry.kind = dpc_consts.ENTRY_KIND_OUTBOUND
        entry.save()
        # Reload the monthly balances
        monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=monthly_balance.pk
        )
        another_source_monthly_balance = dpc_models.MonthlyBalance.objects.get(
            pk=another_source_monthly_balance.pk
        )
        # Check the monthly balances *after* the Entry modification
        self.assertEqual(monthly_balance.end, 100 - 50)
        self.assertEqual(another_source_monthly_balance.end, 101)

    # refs #11704
    def test_all_monthly_balances_update_when_changing_source_of_entry(self):
        """
        Source with 100.00€ monthly balance
        Source with 101.00€ monthly balance
        Entry of 50.00€ INBOUND in the first source, but spanning more than one monthly balance
        Check that the monthly balances are ok
        Move the entry to the other source
        Check that the monthly balances are ok, as expected
        """
        # We need another Source
        another_source = dpc_models.Source.objects.create(name="another_test_source")
        # And MonthlyBalance objects
        for day in [self.month_minus_two, self.month_minus_one, self.day]:
            dpc_models.MonthlyBalance.objects.create(
                date=day,
                source=self.source,
                start=100.00,
                end=100.00,
            )
            dpc_models.MonthlyBalance.objects.create(
                date=day,
                source=another_source,
                start=101.00,
                end=101.00,
            )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.month_minus_one.replace(day=15),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Check the monthly balances
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_two)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_one)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100 + 50)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.day)
        self.assertEqual(mb.start, 100 + 50)
        self.assertEqual(mb.end, 100 + 50)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_two)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_one)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.day)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        # Move the Entry to the other source
        entry.source = another_source
        entry.save()
        # Check the monthly balances *after* the Entry modification
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_two)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_one)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.day)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_two)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_one)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101 + 50)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.day)
        self.assertEqual(mb.start, 101 + 50)
        self.assertEqual(mb.end, 101 + 50)

    # refs #11704
    def test_monthly_balance_update_when_deleting_entry(self):
        """
        Source with 100.00€ monthly balance
        Entry of 50.00€ INBOUND
        Check that the monthly balances is 150€
        Delete the entry
        Check that the monthly balance is back to 100€
        """
        monthly_balance = dpc_models.MonthlyBalance.objects.create(
            date=self.day,
            source=self.source,
            start=100.00,
            end=100.00,
        )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.today,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Check the monthly balance
        monthly_balance = dpc_models.MonthlyBalance.objects.get(pk=monthly_balance.pk)
        self.assertEqual(monthly_balance.start, 100)
        self.assertEqual(monthly_balance.end, 100 + 50)
        # Delete the Entry
        entry.delete()
        # Check the monthly balances *after* the Entry deletion
        monthly_balance = dpc_models.MonthlyBalance.objects.get(pk=monthly_balance.pk)
        self.assertEqual(monthly_balance.start, 100)
        self.assertEqual(monthly_balance.end, 100)

    # refs #11704
    def test_all_monthly_balances_update_when_deleting_entry(self):
        """
        Source with 100.00€ monthly balance
        Source with 101.00€ monthly balance
        Entry of 50.00€ INBOUND in the first source, but spanning more than one monthly balance
        Check that the monthly balances are ok
        Delete the entry
        Check that the monthly balances are ok, as expected
        """
        # We need another Source
        another_source = dpc_models.Source.objects.create(name="another_test_source")
        # And MonthlyBalance objects
        for day in [self.month_minus_two, self.month_minus_one, self.day]:
            dpc_models.MonthlyBalance.objects.create(
                date=day,
                source=self.source,
                start=100.00,
                end=100.00,
            )
            dpc_models.MonthlyBalance.objects.create(
                date=day,
                source=another_source,
                start=101.00,
                end=101.00,
            )
        # Create an Entry for self.source
        entry = dpc_models.Entry(
            company=self.company,
            source=self.source,
            reason="Reason",
            amount=50,
            date=self.month_minus_one.replace(day=15),
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        entry.save()
        # Check the monthly balances
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_two)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_one)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100 + 50)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.day)
        self.assertEqual(mb.start, 100 + 50)
        self.assertEqual(mb.end, 100 + 50)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_two)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_one)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.day)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        # Delete the Entry
        entry.delete()
        # Check the monthly balances *after* the Entry deletion
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_two)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.month_minus_one)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=self.source, date=self.day)
        self.assertEqual(mb.start, 100)
        self.assertEqual(mb.end, 100)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_two)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.month_minus_one)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
        mb = dpc_models.MonthlyBalance.objects.get(source=another_source, date=self.day)
        self.assertEqual(mb.start, 101)
        self.assertEqual(mb.end, 101)
