# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.contrib.auth.models import Permission, User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
import django_pettycash.models as dpc_models
from model_mommy import mommy


class CRUDPermissionsInTemplatesTest(TestCase):
    """
    Class to test that the links/buttons for CRUD are shown only if the user has
    the needed permission.
    """
    def setUp(self):
        self.today = datetime.date.today()
        self.day = datetime.date(
            day=1,
            month=self.today.month,
            year=self.today.year,
        )
        self.client = Client()
        # Create users
        self.user_empty_perms = User(username='user_empty_perms')
        self.user_empty_perms.set_password('password')
        self.user_empty_perms.save()
        self.user_empty_perms.user_permissions.clear()
        for verb in ['add', 'change', 'delete']:
            for model in ['entry', 'monthlybalance', 'source', 'tag']:
                u = User(username="%s_%s" % (verb, model))
                u.set_password('password')
                u.save()
                u.user_permissions.clear()
                u.user_permissions.add(Permission.objects.get(codename="%s_%s" % (verb, model)))
        # Create objects
        self.tag = mommy.make(dpc_models.Tag, name="test tag")
        self.source = mommy.make(dpc_models.Source, name="test source", description="test description")
        self.monthly_balance = mommy.make(dpc_models.MonthlyBalance,
            source=self.source,
            date=self.day,
            start=100,
            end=100,
        )
        self.entry = mommy.make(dpc_models.Entry,
            source=self.source,
            amount=10.00,
            date=self.today,
            reason="test reason",
        )


class CRUDPermissionsInTemplatesForEntryTest(CRUDPermissionsInTemplatesTest):
    def test_entry_list_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_entry, entry_create link
        self.client.login(
            username='add_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertContains(response, reverse('dpc_entry_create'))
        # change_entry, entry_update link
        self.client.login(
            username='change_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertContains(response, reverse('dpc_entry_update', args=[self.entry.pk]))
        # delete_entry, entry_delete link
        self.client.login(
            username='delete_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertContains(response, reverse('dpc_entry_delete', args=[self.entry.pk]))

    def test_entry_detail_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_entry_detail', args=[self.entry.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # change_entry, entry_update link
        self.client.login(
            username='change_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_detail', args=[self.entry.pk]))
        self.assertContains(response, reverse('dpc_entry_update', args=[self.entry.pk]))
        # delete_entry, entry_delete link
        self.client.login(
            username='delete_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_detail', args=[self.entry.pk]))
        self.assertContains(response, reverse('dpc_entry_delete', args=[self.entry.pk]))

    def test_entry_confirm_delete_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_entry_delete', args=[self.entry.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # delete_entry, button
        self.client.login(
            username='delete_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_delete', args=[self.entry.pk]))
        self.assertContains(response, 'id_submit_button')

    def test_entry_create_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_entry_create'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_entry, button
        self.client.login(
            username='add_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_create'))
        self.assertContains(response, 'id_submit_button')

    def test_entry_update_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_entry_update', args=[self.entry.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # change_entry, button
        self.client.login(
            username='change_entry',
            password='password')
        response = self.client.get(reverse('dpc_entry_update', args=[self.entry.pk]))
        self.assertContains(response, 'id_submit_button')


class CRUDPermissionsInTemplatesForSourceTest(CRUDPermissionsInTemplatesTest):
    def test_source_list_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_source_list'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_source, entry_create link
        self.client.login(
            username='add_source',
            password='password')
        response = self.client.get(reverse('dpc_source_list'))
        self.assertContains(response, reverse('dpc_source_create'))
        # change_source, source_update link
        self.client.login(
            username='change_source',
            password='password')
        response = self.client.get(reverse('dpc_source_list'))
        self.assertContains(response, reverse('dpc_source_update', args=[self.source.pk]))
        # delete_source, source_delete link
        self.client.login(
            username='delete_source',
            password='password')
        response = self.client.get(reverse('dpc_source_list'))
        self.assertContains(response, reverse('dpc_source_delete', args=[self.source.pk]))

    def test_source_detail_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_source_detail', args=[self.source.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # change_source, source_update link
        self.client.login(
            username='change_source',
            password='password')
        response = self.client.get(reverse('dpc_source_detail', args=[self.source.pk]))
        self.assertContains(response, reverse('dpc_source_update', args=[self.source.pk]))
        # delete_source, source_delete link
        self.client.login(
            username='delete_source',
            password='password')
        response = self.client.get(reverse('dpc_source_detail', args=[self.source.pk]))
        self.assertContains(response, reverse('dpc_source_delete', args=[self.source.pk]))

    def test_source_confirm_delete_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_source_delete', args=[self.source.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # delete_source, button
        self.client.login(
            username='delete_source',
            password='password')
        response = self.client.get(reverse('dpc_source_delete', args=[self.source.pk]))
        self.assertContains(response, 'id_submit_button')

    def test_source_create_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_source_create'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_source, button
        self.client.login(
            username='add_source',
            password='password')
        response = self.client.get(reverse('dpc_source_create'))
        self.assertContains(response, 'id_submit_button')

    def test_source_update_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_source_update', args=[self.source.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # change_source, button
        self.client.login(
            username='change_source',
            password='password')
        response = self.client.get(reverse('dpc_source_update', args=[self.source.pk]))
        self.assertContains(response, 'id_submit_button')


class CRUDPermissionsInTemplatesForTagTest(CRUDPermissionsInTemplatesTest):
    def test_tag_list_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_tag_list'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_tag, tagcreate link
        self.client.login(
            username='add_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_list'))
        self.assertContains(response, reverse('dpc_tag_create'))
        # change_tag, tag_update link
        self.client.login(
            username='change_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_list'))
        self.assertContains(response, reverse('dpc_tag_update', args=[self.tag.pk]))
        # delete_tag, tag_delete link
        self.client.login(
            username='delete_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_list'))
        self.assertContains(response, reverse('dpc_tag_delete', args=[self.tag.pk]))

    def test_tag_detail_page(self):
        # No perms, no links
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_tag_detail', args=[self.tag.pk]))
        self.assertNotContains(response, reverse('dpc_tag_update', args=[self.tag.pk]))
        self.assertNotContains(response, reverse('dpc_tag_delete', args=[self.tag.pk]))
        self.client.logout()
        # change_tag, tag_update link
        self.client.login(
            username='change_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_detail', args=[self.tag.pk]))
        self.assertContains(response, reverse('dpc_tag_update', args=[self.tag.pk]))
        # delete_tag, tag_delete link
        self.client.login(
            username='delete_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_detail', args=[self.tag.pk]))
        self.assertContains(response, reverse('dpc_tag_delete', args=[self.tag.pk]))

    def test_tag_confirm_delete_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_tag_delete', args=[self.tag.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # delete_tag, button
        self.client.login(
            username='delete_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_delete', args=[self.tag.pk]))
        self.assertContains(response, 'id_submit_button')

    def test_tag_create_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_tag_create'))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # add_tag, button
        self.client.login(
            username='add_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_create'))
        self.assertContains(response, 'id_submit_button')

    def test_tag_update_page(self):
        # No perms, 403
        self.client.login(
            username='user_empty_perms',
            password='password')
        response = self.client.get(reverse('dpc_tag_update', args=[self.tag.pk]))
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # change_tag, button
        self.client.login(
            username='change_tag',
            password='password')
        response = self.client.get(reverse('dpc_tag_update', args=[self.tag.pk]))
        self.assertContains(response, 'id_submit_button')
