# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.conf import settings
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
import django_pettycash.consts as dpcconsts
import django_pettycash.models as dpcmodels


DATEFORMAT = getattr(settings, 'DATEFORMAT')


class ViewsAreRespondingTest(TestCase):
    """
    Class to test that views are at least responding
    """
    def setUp(self):
        self.client = Client()

    # refs #4921
    def test_dpc_stats_balance_by_tag(self):
        response = self.client.get(reverse('dpc_stats_balance_by_tag'))
        self.assertIn(response.status_code, [302, 200])


class PermissionViewsTest(TestCase):
    """
    Class to test the permissions in the views
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 2)
        self.assertFalse(self.user.is_superuser)
        self.assertTrue(self.superuser.is_superuser)

    # refs #4921
    def test_login_required_dpc_stats_balance_by_tag(self):
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_stats_balance_by_tag'))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_stats_balance_by_tag'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_stats_balance_by_tag'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()


class FunctionalityTests(TestCase):
    """
    Class to test the functionality of the views
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()

    def test_stats_balance_by_tag(self):
        # Create Tag0
        tag0 = dpcmodels.Tag.objects.create(name='Tag 0')
        # Create Tag1
        tag1 = dpcmodels.Tag.objects.create(name='Tag 1')
        # Create Tag2
        tag2 = dpcmodels.Tag.objects.create(name='Tag 2')
        # Create Source0
        source0 = dpcmodels.Source.objects.create(
            name='Source0', description='Source 0')
        # Create Source1
        source1 = dpcmodels.Source.objects.create(
            name='Source1', description='Source 1')
        # Create December 2014 MonthlyBalance objects
        mb_201412_0 = dpcmodels.MonthlyBalance.objects.create(
            source=source0,
            date=datetime.date(2014, 12, 1),
            start=1000,
        )
        mb_201412_1 = dpcmodels.MonthlyBalance.objects.create(
            source=source1,
            date=datetime.date(2014, 12, 1),
            start=10000,
        )
        self.assertEqual(dpcmodels.Tag.objects.all().count(), 3)
        self.assertEqual(dpcmodels.Source.objects.all().count(), 2)
        self.assertEqual(dpcmodels.MonthlyBalance.objects.all().count(), 2)
        # Create Entry objects
        # source0, inbound, 10, no tags
        dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 1),
        )
        # source0, outbound, 15, no tags
        dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 2),
        )
        # source1, inbound, 10, no tags
        dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 3),
        )
        # source1, outbound, 15, no tags
        dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 4),
        )
        # source0, inbound, 10, Tag0
        entry = dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 5),
        )
        entry.tags.add(tag0)
        # source0, outbound, 15, Tag0
        entry = dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 6),
        )
        entry.tags.add(tag0)
        # source1, inbound, 10, Tag0
        entry = dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 7),
        )
        entry.tags.add(tag0)
        # source1, outbound, 15, Tag0
        entry = dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 8),
        )
        entry.tags.add(tag0)
        # source0, inbound, 10, Tag0 Tag1
        entry = dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 9),
        )
        entry.tags = [tag0, tag1]
        # source0, outbound, 15, Tag0 Tag1
        entry = dpcmodels.Entry.objects.create(
            source=source0,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 10),
        )
        entry.tags = [tag0, tag1]
        # source1, inbound, 10, Tag0 Tag1
        entry = dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_INBOUND,
            amount=10.0,
            reason="Test",
            date=datetime.date(2014, 12, 11),
        )
        entry.tags = [tag0, tag1]
        # source1, outbound, 15, Tag0 Tag1
        entry = dpcmodels.Entry.objects.create(
            source=source1,
            kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            amount=15.0,
            reason="Test",
            date=datetime.date(2014, 12, 12),
        )
        entry.tags = [tag0, tag1]
        self.assertEqual(dpcmodels.Entry.objects.all().count(), 12)
        mb_201412_0 = dpcmodels.MonthlyBalance.objects.get(pk=mb_201412_0.pk)
        self.assertEqual(mb_201412_0.start, 1000)
        self.assertEqual(mb_201412_0.end, 985)
        mb_201412_1 = dpcmodels.MonthlyBalance.objects.get(pk=mb_201412_1.pk)
        self.assertEqual(mb_201412_1.start, 10000)
        self.assertEqual(mb_201412_1.end, 9985)
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        # Do a GET
        response = self.client.get(reverse('dpc_stats_balance_by_tag'))
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertIn('object_list', ctx)
        # Assert that object_list is empty
        self.assertEqual(len(ctx['object_list']), 0)
        # Do a POST, Tag0, no dates
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': tag0.pk,
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 20)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 30)
        self.assertEqual(ctx['data']['Source0']['total'], -10)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 2)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 2)
        self.assertEqual(ctx['data']['Source0']['count'], 4)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 20)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 30)
        self.assertEqual(ctx['data']['Source1']['total'], -10)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 2)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 2)
        self.assertEqual(ctx['data']['Source1']['count'], 4)
        self.assertEqual(ctx['min_date'], datetime.date(2014, 12, 5))
        self.assertEqual(ctx['max_date'], datetime.date(2014, 12, 12))
        self.assertEqual(ctx['inbound_count'], 4)
        self.assertEqual(ctx['outbound_count'], 4)
        self.assertEqual(ctx['count'], 8)
        self.assertEqual(ctx['inbound_total'], 40)
        self.assertEqual(ctx['outbound_total'], 60)
        self.assertEqual(ctx['total'], -20)
        # Do a POST, Tag0, start date
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': tag0.pk,
            'from_date': datetime.date(2014, 12, 8).strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 15)
        self.assertEqual(ctx['data']['Source0']['total'], -5)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['count'], 2)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 30)
        self.assertEqual(ctx['data']['Source1']['total'], -20)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 2)
        self.assertEqual(ctx['data']['Source1']['count'], 3)
        self.assertEqual(ctx['min_date'], datetime.date(2014, 12, 8))
        self.assertEqual(ctx['max_date'], datetime.date(2014, 12, 12))
        self.assertEqual(ctx['inbound_count'], 2)
        self.assertEqual(ctx['outbound_count'], 3)
        self.assertEqual(ctx['count'], 5)
        self.assertEqual(ctx['inbound_total'], 20)
        self.assertEqual(ctx['outbound_total'], 45)
        self.assertEqual(ctx['total'], -25)
        # Do a POST, Tag0, end date
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': tag0.pk,
            'to_date': datetime.date(2014, 12, 8).strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 15)
        self.assertEqual(ctx['data']['Source0']['total'], -5)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['count'], 2)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 15)
        self.assertEqual(ctx['data']['Source1']['total'], -5)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 1)
        self.assertEqual(ctx['data']['Source1']['count'], 2)
        self.assertEqual(ctx['min_date'], datetime.date(2014, 12, 5))
        self.assertEqual(ctx['max_date'], datetime.date(2014, 12, 8))
        self.assertEqual(ctx['inbound_count'], 2)
        self.assertEqual(ctx['outbound_count'], 2)
        self.assertEqual(ctx['count'], 4)
        self.assertEqual(ctx['inbound_total'], 20)
        self.assertEqual(ctx['outbound_total'], 30)
        self.assertEqual(ctx['total'], -10)
        # Do a POST, Tag1, no dates
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': tag1.pk,
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 15)
        self.assertEqual(ctx['data']['Source0']['total'], -5)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 1)
        self.assertEqual(ctx['data']['Source0']['count'], 2)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 10)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 15)
        self.assertEqual(ctx['data']['Source1']['total'], -5)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 1)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 1)
        self.assertEqual(ctx['data']['Source1']['count'], 2)
        self.assertEqual(ctx['min_date'], datetime.date(2014, 12, 9))
        self.assertEqual(ctx['max_date'], datetime.date(2014, 12, 12))
        self.assertEqual(ctx['inbound_count'], 2)
        self.assertEqual(ctx['outbound_count'], 2)
        self.assertEqual(ctx['count'], 4)
        self.assertEqual(ctx['inbound_total'], 20)
        self.assertEqual(ctx['outbound_total'], 30)
        self.assertEqual(ctx['total'], -10)
        # Do a POST, Tag0 and Tag1, no dates
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': [tag0.pk, tag1.pk],
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 20)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 30)
        self.assertEqual(ctx['data']['Source0']['total'], -10)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 2)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 2)
        self.assertEqual(ctx['data']['Source0']['count'], 4)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 20)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 30)
        self.assertEqual(ctx['data']['Source1']['total'], -10)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 2)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 2)
        self.assertEqual(ctx['data']['Source1']['count'], 4)
        self.assertEqual(ctx['min_date'], datetime.date(2014, 12, 5))
        self.assertEqual(ctx['max_date'], datetime.date(2014, 12, 12))
        self.assertEqual(ctx['inbound_count'], 4)
        self.assertEqual(ctx['outbound_count'], 4)
        self.assertEqual(ctx['count'], 8)
        self.assertEqual(ctx['inbound_total'], 40)
        self.assertEqual(ctx['outbound_total'], 60)
        self.assertEqual(ctx['total'], -20)
        # Do a POST, Tag2, no dates
        response = self.client.post(reverse('dpc_stats_balance_by_tag'), {
            'tags': tag2.pk,
        })
        self.assertEqual(response.status_code, 200)
        ctx = response.context
        self.assertEqual(ctx['data']['Source0']['inbound_total'], 0)
        self.assertEqual(ctx['data']['Source0']['outbound_total'], 0)
        self.assertEqual(ctx['data']['Source0']['total'], 0)
        self.assertEqual(ctx['data']['Source0']['inbound_count'], 0)
        self.assertEqual(ctx['data']['Source0']['outbound_count'], 0)
        self.assertEqual(ctx['data']['Source0']['count'], 0)
        self.assertEqual(ctx['data']['Source1']['inbound_total'], 0)
        self.assertEqual(ctx['data']['Source1']['outbound_total'], 0)
        self.assertEqual(ctx['data']['Source1']['total'], 0)
        self.assertEqual(ctx['data']['Source1']['inbound_count'], 0)
        self.assertEqual(ctx['data']['Source1']['outbound_count'], 0)
        self.assertEqual(ctx['data']['Source1']['count'], 0)
        self.assertEqual(ctx['min_date'], '')
        self.assertEqual(ctx['max_date'], '')
        self.assertEqual(ctx['inbound_count'], 0)
        self.assertEqual(ctx['outbound_count'], 0)
        self.assertEqual(ctx['count'], 0)
        self.assertEqual(ctx['inbound_total'], 0)
        self.assertEqual(ctx['outbound_total'], 0)
        self.assertEqual(ctx['total'], 0)
