# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Permission, User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
from model_mommy import mommy
from django_pettycash import consts as dpc_consts
from django_pettycash.models import Entry, MonthlyBalance, Source, Tag


DATEFORMAT = getattr(settings, 'DATEFORMAT')
COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')
_today = datetime.date.today()
today_month = _today.strftime("%Y-%m")
previous_month = (_today - datetime.timedelta(days=(_today.day + 1))).strftime("%Y-%m")
Company = apps.get_model(
    app_label=COMPANY_MODEL.split('.')[0],
    model_name=COMPANY_MODEL.split('.')[1],
)


class EntryViewsAreRespondingTest(TestCase):
    """
    Class to test that views related to Entry CRUD are at least responding
    """
    def setUp(self):
        self.client = Client()

    def test_dpc_entry_create(self):
        response = self.client.get(reverse('dpc_entry_create'))
        self.assertIn(response.status_code, [403, 302, 200])

    def test_dpc_entry_list(self):
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertIn(response.status_code, [403, 302, 200])

    def test_dpc_entry_csv_list(self):
        response = self.client.get(reverse('dpc_entry_csv_list'))
        self.assertIn(response.status_code, [302, 200])

    def test_dpc_entry_update(self):
        day = datetime.date(day=1, month=1, year=2015)
        mb = mommy.make(MonthlyBalance, date=day)
        e = mommy.make(Entry, date=day, source=mb.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        response = self.client.get(reverse('dpc_entry_update', kwargs={'pk': e.pk}))
        self.assertIn(response.status_code, [302, 200])
        Entry.objects.all().delete()

    def test_dpc_entry_delete_get(self):
        day = datetime.date(day=1, month=1, year=2015)
        mb = mommy.make(MonthlyBalance, date=day)
        e = mommy.make(Entry, date=day, source=mb.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        response = self.client.get(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        self.assertIn(response.status_code, [302, 200])
        Entry.objects.all().delete()

    def test_dpc_entry_delete_post(self):
        day = datetime.date(day=1, month=1, year=2015)
        mb = mommy.make(MonthlyBalance, date=day)
        e = mommy.make(Entry, date=day, source=mb.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        response = self.client.post(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        self.assertIn(response.status_code, [302, 200])
        Entry.objects.all().delete()

    def test_dpc_entry_detail(self):
        day = datetime.date(day=1, month=1, year=2015)
        mb = mommy.make(MonthlyBalance, date=day)
        e = mommy.make(Entry, date=day, source=mb.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': e.pk}))
        self.assertIn(response.status_code, [302, 200])
        Entry.objects.all().delete()


class PermissionEntryViewsTest(TestCase):
    """
    Class to test the permissions in the views related to Entry CRUD
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 2)
        self.assertFalse(self.user.is_superuser)
        self.assertTrue(self.superuser.is_superuser)
        # Create objects
        self.today = datetime.date.today()
        self.day = datetime.date(
            day=1,
            month=self.today.month,
            year=self.today.year,
        )
        self.source = mommy.make(Source)
        self.monthly_balance = mommy.make(MonthlyBalance,
            source=self.source,
            date=self.day,
            start=100.0,
        )
        self.monthly_balance.update_balance()

    def test_permission_dpc_entry_list(self):
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_entry_list'))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user, no permissions...
        self.user.user_permissions.clear()
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_list'))
        # ...403
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # Logged as user, add_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, change_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, delete_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_entry_create(self):
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_entry_create'))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_create'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_entry_create'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_entry_update_get(self):
        # Create one Entry object
        mommy.make(Entry, date=self.today, source=self.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_entry_update', kwargs={'pk': entries[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_update', kwargs={'pk': entries[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_entry_update', kwargs={'pk': entries[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_entry_update_post(self):
        amount = 1.00
        # Create one Entry object
        mommy.make(Entry, date=self.today, source=self.source, amount=amount)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        e = entries[0]
        self.assertEqual(e.amount, amount)
        # Not logged in...
        self.client.logout()
        cpk = e.company.pk if e.company else ''
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': e.pk}), {
            'pk': e.pk,
            'company': cpk,
            'source': e.source.pk,
            'kind': e.kind,
            'amount': amount * 2,
            'reason': e.reason,
            'date': e.date,
        })
        # ...302
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Entry.objects.get(pk=e.pk).amount, amount)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_entry'))
        self.client.login(username='user', password='user')
        cpk = e.company.pk if e.company else ''
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': e.pk}), {
            'pk': e.pk,
            'company': cpk,
            'source': e.source.pk,
            'kind': e.kind,
            'amount': amount * 2,
            'reason': e.reason,
            'date': e.date.strftime(DATEFORMAT),
        })
        # ...good 302
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        self.assertEqual(Entry.objects.get(pk=e.pk).amount, amount * 2)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        cpk = e.company.pk if e.company else ''
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': e.pk}), {
            'pk': e.pk,
            'company': cpk,
            'source': e.source.pk,
            'kind': e.kind,
            'amount': amount * 3,
            'reason': e.reason,
            'date': e.date.strftime(DATEFORMAT),
        })
        # ...good 302
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        self.assertEqual(Entry.objects.get(pk=e.pk).amount, amount * 3)
        self.client.logout()

    def test_permission_dpc_entry_delete_get(self):
        # Create one Entry object
        e = mommy.make(Entry, date=self.today, source=self.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        Entry.objects.all().delete()
        self.client.logout()

    def test_permission_dpc_entry_delete_post(self):
        # Create one Entry object
        e = mommy.make(Entry, date=self.today, source=self.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.post(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_entry'))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, 'http://testserver' + reverse('dpc_entry_list'))
        self.assertEqual(entries.count(), 0)
        self.client.logout()
        # Create one Entry object
        e = mommy.make(Entry, date=self.today, source=self.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.post(reverse('dpc_entry_delete', kwargs={'pk': e.pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, 'http://testserver' + reverse('dpc_entry_list'))
        self.assertEqual(entries.count(), 0)
        Entry.objects.all().delete()
        self.client.logout()

    def test_permission_dpc_entry_detail(self):
        # Create one Entry object
        mommy.make(Entry, date=self.today, source=self.source, amount=1.00)
        entries = Entry.objects.all()
        self.assertEqual(entries.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': entries[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user, no permissions...
        self.user.user_permissions.clear()
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': entries[0].pk}))
        # ...403
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # Logged as user, add_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': entries[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, change_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': entries[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, delete_entry permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_entry'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_entry_detail', kwargs={'pk': entries[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()


class ViewsTest(TestCase):
    def setUp(self):
        self.client = Client()
        self.user0 = User(
            username='user0',
            is_superuser=False,
        )
        self.user0.set_password('user0')
        self.user0.save()
        self.user1 = User(
            username='user1',
            is_superuser=False,
        )
        self.user1.set_password('user1')
        self.user1.save()
        self.user42 = User(
            username='user42',
            is_superuser=False,
        )
        self.user42.set_password('user42')
        self.user42.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 4)
        self.assertFalse(self.user0.is_superuser)
        self.assertFalse(self.user1.is_superuser)
        self.assertTrue(self.superuser.is_superuser)
        # Create objects
        self.today = datetime.date.today()
        self.day = datetime.date(
            day=1,
            month=self.today.month,
            year=self.today.year,
        )
        self.company = mommy.make(Company)
        self.source = mommy.make(Source)
        self.tag = mommy.make(Tag)
        self.monthly_balance = mommy.make(MonthlyBalance,
            source=self.source,
            date=self.day,
        )

    def test_created_by_is_saved_when_creating_entry(self):
        self.user0.user_permissions.clear()
        self.user0.user_permissions.add(Permission.objects.get(codename="add_entry"))
        self.client.login(username='user0', password='user0')
        how_many_entries = Entry.objects.all().count()
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 100,
            'reason': "Reason",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        entries = Entry.objects.all().order_by('-created')
        entry = entries[0]
        self.assertEqual(entries.count(), how_many_entries + 1)
        self.assertEqual(entry.created_by, self.user0)
        self.assertEqual(entry.modified_by, self.user0)

    def test_modified_by_is_saved_when_editing_entry(self):
        self.user0.user_permissions.clear()
        self.user0.user_permissions.add(Permission.objects.get(codename="add_entry"))
        # Create one new Entry as user0
        logged_in = self.client.login(username='user0', password='user0')
        self.assertTrue(logged_in)
        how_many_initial_entries = Entry.objects.all().count()
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': self.company.pk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 100,
            'reason': "Reason",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        # Check that the test created one Entry
        entries = Entry.objects.all().order_by('-created')
        entry = entries[0]
        self.assertEqual(entries.count(), how_many_initial_entries + 1)
        # Check the fields
        self.assertEqual(entry.created_by, self.user0)
        self.assertEqual(entry.modified_by, self.user0)
        self.assertEqual(entry.amount, 100)
        self.client.logout()
        # Edit the last entry created, as user1
        self.user1.user_permissions.clear()
        self.user1.user_permissions.add(Permission.objects.get(codename="change_entry"))
        logged_in = self.client.login(username='user1', password='user1')
        self.assertTrue(logged_in)
        response = self.client.post(reverse('dpc_entry_update', kwargs={'pk': entry.pk}), {
            'company': self.company.pk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            # Change the amount
            'amount': 120,
            'reason': "Reason",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        entries = Entry.objects.all().order_by('-created')
        # Check that there is no new entry
        self.assertEqual(entries.count(), how_many_initial_entries + 1)
        # Check the fields
        entry = entries[0]
        self.assertEqual(entry.amount, 120)
        self.assertEqual(entry.created_by, self.user0)
        self.assertEqual(entry.modified_by, self.user1)

    def test_entry_creation_without_any_monthly_balance_for_the_source(self):
        # Ensure to delete all MonthlyBalance objects
        self.monthly_balance = None
        MonthlyBalance.objects.all().delete()
        # Create another Source and a MonthlyBalance for the new Source
        another_source = mommy.make(Source)
        mommy.make(MonthlyBalance,
            source=another_source,
            date=self.day,
            start=100,
            end=100,
        )
        # Logged as user...
        self.user42.user_permissions.clear()
        self.user42.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.user42.user_permissions.add(Permission.objects.get(codename='add_monthlybalance'))
        self.client.login(username='user42', password='user42')
        cpk = self.company.pk if self.company else ''
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': cpk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 101.00,
            'reason': "Test reason 42_42",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['form'].errors), 1)

    def test_entry_creation_without_previous_nor_current_month_monthly_balance_for_the_source(self):
        # Ensure to delete all MonthlyBalance objects
        self.monthly_balance = None
        MonthlyBalance.objects.all().delete()
        # Create a MonthlyBalance two months in the past wrt today
        two_months_before_today = (self.day - datetime.timedelta(days=35)).replace(day=1)
        mommy.make(MonthlyBalance,
            source=self.source,
            date=two_months_before_today,
            start=100,
            end=100,
        )
        entry_pks_before = [x.pk for x in Entry.objects.all()]
        # Logged as user...
        self.user42.user_permissions.clear()
        self.user42.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.user42.user_permissions.add(Permission.objects.get(codename='add_monthlybalance'))
        self.client.login(username='user42', password='user42')
        cpk = self.company.pk if self.company else ''
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': cpk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 100.00,
            'reason': "Test reason 42_42",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': self.today.strftime("%Y-%m")}),
        )
        # Check also objects
        new_entry = Entry.objects.get(
            company=self.company,
            source=self.source,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
            amount=100,
            reason="Test reason 42_42",
            date=self.today,
        )
        self.assertNotIn(new_entry.pk, entry_pks_before)
        monthly_balances = MonthlyBalance.objects.filter(source=self.source)
        self.assertEqual(monthly_balances.count(), 3)
        # One MonthlyBalance two months before today
        two_months_before = monthly_balances.get(date=two_months_before_today)
        self.assertEqual(two_months_before.start, 100)
        self.assertEqual(two_months_before.end, 100)
        # One MonthlyBalance in the previous month
        previous_month = monthly_balances.get(date=(self.day - datetime.timedelta(days=1)).replace(day=1))
        self.assertEqual(previous_month.start, 100)
        self.assertEqual(previous_month.end, 100)
        # One MonthlyBalance in the current month
        current_month = monthly_balances.get(date=self.day)
        self.assertEqual(current_month.start, 100)
        self.assertEqual(current_month.end, 200)

    def test_entry_creation_without_previous_month_monthly_balance(self):
        # Ensure to delete all MonthlyBalance objects
        self.monthly_balance = None
        MonthlyBalance.objects.all().delete()
        self.assertEqual(MonthlyBalance.objects.all().count(), 0)
        # Create the current month MonthlyBalance object
        self.curr_mb = mommy.make(MonthlyBalance,
            source=self.source,
            date=self.day,
            start=100,
            end=100,
        )
        # Logged as user...
        self.user42.user_permissions.clear()
        self.user42.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user42', password='user42')
        cpk = self.company.pk if self.company else ''
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': cpk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 101.00,
            'reason': "Test reason 42_42",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        entry = Entry.objects.get(reason="Test reason 42_42")
        self.curr_mb = MonthlyBalance.objects.get(
            source=self.source,
            date=self.day,
        )
        self.assertEqual(MonthlyBalance.objects.all().count(), 1)
        self.assertEqual(entry.amount, 101)
        self.assertEqual(self.curr_mb.start, 100)
        self.assertEqual(self.curr_mb.end, 201)

    def test_entry_creation_without_current_month_monthly_balance(self):
        # Ensure to delete all MonthlyBalance objects
        self.monthly_balance = None
        MonthlyBalance.objects.all().delete()
        self.assertEqual(MonthlyBalance.objects.all().count(), 0)
        # Create the previous month MonthlyBalance object
        day = self.day - datetime.timedelta(days=1)
        first_day_previous_month = datetime.date(
            day=1,
            month=day.month,
            year=day.year,
        )
        mommy.make(MonthlyBalance,
            source=self.source,
            date=first_day_previous_month,
            start=100,
            end=100,
        )
        # Logged as user...
        self.user42.user_permissions.clear()
        self.user42.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user42', password='user42')
        cpk = self.company.pk if self.company else ''
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': cpk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 101.00,
            'reason': "Test reason 42_42",
            'date': self.today.strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': today_month}),
        )
        # Check that the entry is created
        entry = Entry.objects.get(reason="Test reason 42_42")
        # Get current month monthly balance
        self.curr_mb = MonthlyBalance.objects.get(
            source=self.source,
            date=self.day,
        )
        # Get previous month monthly balance
        self.prev_mb = MonthlyBalance.objects.get(
            source=self.source,
            date=first_day_previous_month,
        )
        # Check that there are two MonthlyBalance objects
        self.assertEqual(MonthlyBalance.objects.all().count(), 2)
        self.assertEqual(entry.amount, 101)
        # Check that previous month monthly balance has the data we created with
        self.assertEqual(self.prev_mb.start, 100)
        self.assertEqual(self.prev_mb.end, 100)
        # Check that current month monthly balance is created automatically with the right data
        self.assertEqual(self.curr_mb.start, 100)
        self.assertEqual(self.curr_mb.end, 201)

    def test_monthly_balance_cascading_update_when_saving_entry(self):
        # Ensure to delete all MonthlyBalance objects
        self.monthly_balance = None
        MonthlyBalance.objects.all().delete()
        self.assertEqual(MonthlyBalance.objects.all().count(), 0)
        # Create the current month MonthlyBalance object
        cm = mommy.make(MonthlyBalance,
            source=self.source,
            date=self.day,
            start=100,
            end=100,
        )
        # Create the previous month MonthlyBalance object
        day = self.day - datetime.timedelta(days=1)
        first_day_previous_month = datetime.date(
            day=1,
            month=day.month,
            year=day.year,
        )
        pm = mommy.make(MonthlyBalance,
            source=self.source,
            date=first_day_previous_month,
            start=100,
            end=100,
        )
        # Create the previous previous month MonthlyBalance object
        day = first_day_previous_month - datetime.timedelta(days=1)
        first_day_previous_previous_month = datetime.date(
            day=1,
            month=day.month,
            year=day.year,
        )
        ppm = mommy.make(MonthlyBalance,
            source=self.source,
            date=first_day_previous_previous_month,
            start=100,
            end=100,
        )
        # Create an Entry in the previous month
        self.user42.user_permissions.clear()
        self.user42.user_permissions.add(Permission.objects.get(codename='add_entry'))
        self.client.login(username='user42', password='user42')
        cpk = self.company.pk if self.company else ''
        response = self.client.post(reverse('dpc_entry_create'), {
            'company': cpk,
            'source': self.source.pk,
            'kind': dpc_consts.ENTRY_KIND_INBOUND,
            'amount': 10.00,
            'reason': "Test reason 42_42",
            'date': (first_day_previous_month + datetime.timedelta(days=10)).strftime(DATEFORMAT),
        })
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={'month': previous_month}),
        )
        # Check that monthly balances are updated, when needed
        # Reload objects from database
        cm = MonthlyBalance.objects.get(pk=cm.pk)
        pm = MonthlyBalance.objects.get(pk=pm.pk)
        ppm = MonthlyBalance.objects.get(pk=ppm.pk)
        # previous previous month is untouched
        self.assertEqual(ppm.start, 100)
        self.assertEqual(ppm.end, 100)
        # previous month has start untouched, end updated
        self.assertEqual(pm.start, 100)
        self.assertEqual(pm.end, 110)
        # current month has start and end updated
        self.assertEqual(cm.start, 110)
        self.assertEqual(cm.end, 110)


class EntryViewsBugReportsTest(TestCase):
    def setUp(self):
        self.client = Client()
        self.user0 = User(
            username='user0',
            is_superuser=False,
        )
        self.user0.set_password('user0')
        self.user0.save()
        self.user1 = User(
            username='user1',
            is_superuser=False,
        )
        self.user1.set_password('user1')
        self.user1.save()
        self.user42 = User(
            username='user42',
            is_superuser=False,
        )
        self.user42.set_password('user42')
        self.user42.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 4)
        self.assertFalse(self.user0.is_superuser)
        self.assertFalse(self.user1.is_superuser)
        self.assertTrue(self.superuser.is_superuser)
        # Create objects
        self.today = datetime.date.today()
        self.day = self.today.replace(day=1)
        self.company = mommy.make(Company)
        self.source = mommy.make(Source)
        self.tag = mommy.make(Tag)
        self.monthly_balance = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=self.day,
            start=100,
        )

    # refs #5523
    def test_entry_modification_that_makes_it_the_first_entry_in_the_previous_month_without_balance(self):
        # Assert that there is one MonthlyBalance for the current month
        self.assertEqual(MonthlyBalance.objects.filter(date=self.day).count(), 1)
        # Assert that there are no MonthlyBalance for the previous month
        first_day_previous_month = (self.day - datetime.timedelta(days=1)).replace(day=1)
        self.assertEqual(MonthlyBalance.objects.filter(date=first_day_previous_month).count(), 0)
        # Create an Entry in the current month
        this_month_entry = mommy.make(
            Entry,
            company=self.company,
            source=self.source,
            date=self.today,
            amount=10,
        )
        # Edit this entry and change the date to be in previous month
        self.user1.user_permissions.clear()
        self.user1.user_permissions.add(Permission.objects.get(codename="change_entry"))
        self.user1.user_permissions.add(Permission.objects.get(codename="add_monthlybalance"))
        logged_in = self.client.login(username='user1', password='user1')
        self.assertTrue(logged_in)
        self.assertEqual(this_month_entry.date, self.today)
        response = self.client.post(
            reverse('dpc_entry_update', kwargs={'pk': this_month_entry.pk}), {
                'company': self.company.pk,
                'source': self.source.pk,
                'kind': this_month_entry.kind,
                'amount': 10,
                'reason': this_month_entry.reason,
                # Change the date
                'date': first_day_previous_month.strftime(DATEFORMAT),
            },
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
            'http://testserver' + reverse('dpc_entry_list', kwargs={
                'month': first_day_previous_month.strftime("%Y-%m")
            }),
        )
        this_month_entry = Entry.objects.get(pk=this_month_entry.pk)
        self.assertEqual(this_month_entry.date, first_day_previous_month)

    # refs #5523
    def test_entry_modification_that_makes_it_the_first_entry_in_the_next_month_without_balance(self):
        # Assert that there is one MonthlyBalance for the current month
        self.assertEqual(MonthlyBalance.objects.filter(date=self.day).count(), 1)
        # Assert that there are no MonthlyBalance for the next month
        first_day_next_month = (self.day + datetime.timedelta(days=35)).replace(day=1)
        self.assertEqual(MonthlyBalance.objects.filter(date=first_day_next_month).count(), 0)
        # Create an Entry in the current month
        this_month_entry = mommy.make(
            Entry,
            company=self.company,
            source=self.source,
            date=self.day,
            amount=10,
        )
        # Edit this entry and change the date to be in next month
        self.user1.user_permissions.clear()
        self.user1.user_permissions.add(Permission.objects.get(codename="change_entry"))
        logged_in = self.client.login(username='user1', password='user1')
        self.assertTrue(logged_in)
        response = self.client.post(
            reverse('dpc_entry_update', kwargs={'pk': this_month_entry.pk}), {
                'company': self.company.pk,
                'source': self.source.pk,
                'kind': this_month_entry.kind,
                'amount': 10,
                'reason': this_month_entry.reason,
                # Change the date
                'date': first_day_next_month.strftime(DATEFORMAT),
            },
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(
            response,
           'http://testserver' + reverse('dpc_entry_list', kwargs={
               'month': first_day_next_month.strftime("%Y-%m"),
           }),
        )

    # refs #11885
    def test_check_autorecreation_of_the_missing_current_month_monthlybalance(self):
        one_month_back_day = (self.day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_one_month_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=one_month_back_day,
            start=100,
        )
        two_months_back_day = (one_month_back_day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_two_months_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=two_months_back_day,
            start=100,
        )
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        logged_in = self.client.login(username='superuser', password='superuser')
        self.assertTrue(logged_in)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        # Create an Entry in the current month
        this_month_entry = mommy.make(
            Entry,
            company=self.company,
            source=self.source,
            date=self.day,
            amount=100,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        # Delete the current MonthlyBalance
        self.monthly_balance.delete()
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 2)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).start,
            100,
        )
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).end,
            100 + 100,
        )

    # refs #11885
    def test_check_autorecreation_of_the_two_missing_monthlybalances(self):
        one_month_back_day = (self.day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_one_month_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=one_month_back_day,
            start=100,
        )
        two_months_back_day = (one_month_back_day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_two_months_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=two_months_back_day,
            start=100,
        )
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        logged_in = self.client.login(username='superuser', password='superuser')
        self.assertTrue(logged_in)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        # Create an Entry in the current month
        this_month_entry = mommy.make(
            Entry,
            company=self.company,
            source=self.source,
            date=self.day.replace(day=2),
            amount=100,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        # Delete the last two MonthlyBalances
        self.monthly_balance.delete()
        monthly_balance_one_month_back.delete()
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 1)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).start,
            100,
        )
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).end,
            100 + 100,
        )

    # refs #11885
    def test_check_autorecreation_of_the_missing_current_month_monthlybalance_when_there_are_two_sources(self):
        another_source = mommy.make(Source, name="Another source")
        mommy.make(
            MonthlyBalance,
            source=another_source,
            date=self.day,
            start=100,
        )
        one_month_back_day = (self.day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_one_month_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=one_month_back_day,
            start=100,
        )
        mommy.make(
            MonthlyBalance,
            source=another_source,
            date=one_month_back_day,
            start=100,
        )
        two_months_back_day = (one_month_back_day - datetime.timedelta(days=1)).replace(day=1)
        monthly_balance_two_months_back = mommy.make(
            MonthlyBalance,
            source=self.source,
            date=two_months_back_day,
            start=100,
        )
        (one_month_back_day - datetime.timedelta(days=1)).replace(day=1)
        mommy.make(
            MonthlyBalance,
            source=another_source,
            date=two_months_back_day,
            start=100,
        )
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        self.assertEqual(MonthlyBalance.objects.filter(source=another_source).count(), 3)
        logged_in = self.client.login(username='superuser', password='superuser')
        self.assertTrue(logged_in)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        # Create an Entry in the current month
        this_month_entry = mommy.make(
            Entry,
            company=self.company,
            source=self.source,
            date=self.day,
            amount=100,
            kind=dpc_consts.ENTRY_KIND_INBOUND,
        )
        # Delete the current MonthlyBalance
        self.monthly_balance.delete()
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 2)
        self.assertEqual(MonthlyBalance.objects.filter(source=another_source).count(), 3)
        response = self.client.get(reverse('dpc_entry_list'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(MonthlyBalance.objects.filter(source=self.source).count(), 3)
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).start,
            100,
        )
        self.assertEqual(
            MonthlyBalance.objects.get(source=self.source, date=self.day).end,
            100 + 100,
        )
        self.assertEqual(MonthlyBalance.objects.filter(source=another_source).count(), 3)
