# coding: utf-8
# Copyright (c) 2015, 2018, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
from django.contrib.auth.models import Permission, User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client
from model_mommy import mommy
from django_pettycash.models import Source, MonthlyBalance


class SourceViewsAreRespondingTest(TestCase):
    """
    Class to test that views related to Source CRUD are at least responding
    """
    def setUp(self):
        self.client = Client()

    def test_dpc_source_create(self):
        response = self.client.get(reverse('dpc_source_create'))
        self.assertIn(response.status_code, [302, 200])

    def test_dpc_source_list(self):
        response = self.client.get(reverse('dpc_source_list'))
        self.assertIn(response.status_code, [403, 302, 200])

    def test_dpc_source_update(self):
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        response = self.client.get(reverse('dpc_source_update', kwargs={'pk': sources[0].pk}))
        self.assertIn(response.status_code, [302, 200])
        Source.objects.all().delete()

    def test_dpc_source_delete_get(self):
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        response = self.client.get(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        self.assertIn(response.status_code, [302, 200])
        Source.objects.all().delete()

    def test_dpc_source_delete_post(self):
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        response = self.client.post(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        self.assertIn(response.status_code, [302, 200])
        Source.objects.all().delete()

    def test_dpc_source_detail(self):
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        self.assertIn(response.status_code, [302, 200])
        Source.objects.all().delete()


class PermissionSourceViewsTest(TestCase):
    """
    Class to test the permissions in the views related to Source CRUD
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 2)
        self.assertFalse(self.user.is_superuser)
        self.assertTrue(self.superuser.is_superuser)

    def test_permission_dpc_source_list(self):
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_source_list'))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user, no permissions...
        self.user.user_permissions.clear()
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_list'))
        # ...403
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # Logged as user, add_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, change_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, delete_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_list'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_source_create(self):
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_source_create'))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_create'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_source_create'))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_source_update(self):
        # Create one Source object
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_source_update', kwargs={'pk': sources[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_update', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_source_update', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_permission_dpc_source_delete_get(self):
        # Create one Source object
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.get(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        Source.objects.all().delete()
        self.client.logout()

    def test_permission_dpc_source_delete_post(self):
        # Create one Source object
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.post(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_source'))
        self.client.login(username='user', password='user')
        response = self.client.post(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, 'http://testserver' + reverse('dpc_source_list'))
        self.assertEqual(sources.count(), 0)
        self.client.logout()
        # Create one Source object
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        # Logged as superuser...
        self.client.login(username='superuser', password='superuser')
        response = self.client.post(reverse('dpc_source_delete', kwargs={'pk': sources[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, 'http://testserver' + reverse('dpc_source_list'))
        self.assertEqual(sources.count(), 0)
        Source.objects.all().delete()
        self.client.logout()

    def test_permission_dpc_source_detail(self):
        # Create one Source object
        mommy.make(Source)
        sources = Source.objects.all()
        self.assertEqual(sources.count(), 1)
        # Not logged in...
        self.client.logout()
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        # ...302
        self.assertEqual(response.status_code, 302)
#        self.assertRedirects(response, 'http://testserver' + reverse('login'))
        self.client.logout()
        # Logged as user, no permissions...
        self.user.user_permissions.clear()
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        # ...403
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        # Logged as user, add_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='add_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, change_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='change_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()
        # Logged as user, delete_source permission...
        self.user.user_permissions.clear()
        self.user.user_permissions.add(Permission.objects.get(codename='delete_source'))
        self.client.login(username='user', password='user')
        response = self.client.get(reverse('dpc_source_detail', kwargs={'pk': sources[0].pk}))
        # ...200
        self.assertEqual(response.status_code, 200)
        self.client.logout()


class SourceViewsTest(TestCase):
    """
    Class to test the behavious of the views related to Source CRUD
    """
    def setUp(self):
        self.client = Client()
        self.user = User(
            username='user',
            is_superuser=False,
        )
        self.user.set_password('user')
        self.user.save()
        self.superuser = User(
            username='superuser',
            is_superuser=True,
        )
        self.superuser.set_password('superuser')
        self.superuser.save()
        self.assertEqual(User.objects.all().count(), 2)
        self.assertFalse(self.user.is_superuser)
        self.assertTrue(self.superuser.is_superuser)

    # refs #11813
    def test_dpc_source_create_creates_also_monthly_balance(self):
        self.assertEqual(Source.objects.all().count(), 0)
        self.assertEqual(MonthlyBalance.objects.all().count(), 0)
        self.client.login(username='superuser', password='superuser')
        response = self.client.post(
            reverse('dpc_source_create'),
            {
                'name': 'Test Source',
                'description': 'This is a test Source',
                'start_balance': 500,
            },
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Source.objects.all().count(), 1)
        self.assertEqual(MonthlyBalance.objects.all().count(), 1)
        source = Source.objects.first()
        monthly_balance = MonthlyBalance.objects.first()
        self.assertRedirects(response, reverse('dpc_source_list'))
        self.assertEqual(monthly_balance.source, source)
        self.assertEqual(monthly_balance.date, datetime.date.today().replace(day=1))
        self.assertEqual(monthly_balance.start, 500)
        self.assertEqual(monthly_balance.end, 500)
        self.client.logout()
