# coding: utf-8
# Copyright (c) 2015, Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import calendar
import codecs
from collections import OrderedDict
from io import BytesIO
import csv
import datetime
from decimal import Decimal


def monthrange(day):
    """
    Given a datetime.date object, it returns a tuple of datetime.date objects like this:
    (first_day_of_month, last_day_of_month)
    """
    weekday_first_day, number_days = calendar.monthrange(day.year, day.month)
    first_day = datetime.date(
        day=1,
        month=day.month,
        year=day.year,
    )
    last_day = datetime.date(
        day=number_days,
        month=day.month,
        year=day.year,
    )
    return (first_day, last_day)


def get_previous_month(day):
    """
    Given a datetime.date object, it returns previous month first day's datetime.date object.
    """
    first_day_current_month = datetime.date(
        day=1,
        month=day.month,
        year=day.year,
    )
    last_day_previous_month = first_day_current_month - datetime.timedelta(days=1)
    return datetime.date(
        day=1,
        month=last_day_previous_month.month,
        year=last_day_previous_month.year,
    )


def get_next_month(day):
    """
    Given a datetime.date object, it returns next month first day's datetime.date object.
    """
    (first_day_current_month, last_day_current_month) = monthrange(day)
    return last_day_current_month + datetime.timedelta(days=1)


def is_month_in_the_future(month):
    """
    Given a datetime.date object, returns True/False depending on the fact that it
    represent a month in the future with respect to today.
    """
    today = datetime.date.today()
    this_month = datetime.date(
        day=1,
        month=today.month,
        year=today.year,
    )
    return month > this_month


def month_string_to_date(month):
    """
    Given a string in 'YYYY-MM' format, returns the datetime.date object of the month's first day
    """
    month = month.split('-')
    this_month = datetime.date(
        day=1,
        month=int(month[1]),
        year=int(month[0]),
    )
    return this_month


def get_entry_list_table_data(day, entries, request):
    """
    Given a datetime.date object and the queryset of entries,
    it returns the data to be rendered in the DPCEntryListView table.
    This is the format of the data returned by this function:
    data = {
        'month': string like 'January 2015',
        'sources': QuerySet with all the Source objects,
        'balances': QuerySet with all the MonthlyBalance objects for this month,
        'entries': [
            {
                'entry': Entry object (to have access to all the related objects),
                'amount_cells': List of empty strings and one cell with entry.amount, ready
                                to be rendered (the amount goes in the right column with respect
                                to source and inbound/outbound).
            }
            [...]
        ],
        'per_source': {
            source.pk: {
                'bpk': monthly_balance.pk,
                'spk': source.pk,
                'name': source.name,
                'start': monthly_balance.start (can be negative),
                'end': monthly_balance.end (can be negative),
                'inbound': sum of the inbound entries' amounts for this source (always positive),
                'outbound': sum of the outbound entries' amounts for this source (always positive),
                'total': inbound - outbound (can be negative),
            },
            [...],
        },
        'global': {
            'start': sum of sources' monthly_balance start amounts (can be negative),
            'end': sum of sources' monthly_balance end amounts (can be negative),
            'inbound': sum of the inbound entries' amounts (always positive),
            'outbound': total of the outbound entries' amounts (always positive),
            'total': inbound - outbound (can be negative),
        },
    }
    """
    from . import models as dpcmodels
    from . import consts as dpcconsts
    data = {
        'month': "%s" % day.strftime("%B %Y"),
        # refs #5087 Use OrderedDict to keep the sorting order
        'per_source': OrderedDict(),
        'global': {
            'start': Decimal('0.00'),
            'end': Decimal('0.00'),
            'inbound': Decimal('0.00'),
            'outbound': Decimal('0.00'),
            'total': Decimal('0.00'),
        },
    }
    # Get sources, sorted by name
    sources_qs = dpcmodels.Source.objects.all().order_by('name')
    # Get balances, sorted by source name
    balances_qs = dpcmodels.MonthlyBalance.objects.filter(
        date__month=day.month,
        date__year=day.year,
    ).order_by('source__name')
    if balances_qs.count() != sources_qs.count():
        # Manca il monthly balance del mese in esame.
        # Prendi tutti i mb precedenti al mese in esame, ordinali per data, e prendi l'ultimo
        # disponibile.
        # Crea tutti i mb che mancano, dall'ultimo esistente a quello che serve, ovviamente
        # ricalcolando i saldi usando le entry che ci sono.
        # Casino? Sì. Possibile collo di bottiglia? Sì. Ma si spera che non succeda quasi mai.
        # E se il mb mancante è il primo? Cioè, se non esistono altri mb precedenti? come si fa a
        # sapere il saldo iniziale per fare tutti i conti? Pensare se sia il caso di aggiungere
        # un controllo sul delete del Monthly Balance per evitare di far cancellare il primo mb
        # della storia di una Source.
        for source in sources_qs.all():
            balance = dpcmodels.MonthlyBalance.objects\
                .filter(date__lte=day.replace(day=1))\
                .filter(source=source)\
                .order_by("date")\
                .last()
            if balance:
                next_month = (balance.date.replace(day=1) + datetime.timedelta(days=32)).replace(day=1)
                while next_month < day.replace(day=1):
                    # You cannot use get_or_create() because if the MB already exists you don't know
                    # the value of start (which is REQUIRED), and if it does not exist you have to
                    # pass it (set to 0, because the MBs will be updated below).
                    # So, try/except it is.
                    try:
                        dpcmodels.MonthlyBalance.objects.get(
                            date=next_month,
                            source=source,
                        )
                    except dpcmodels.MonthlyBalance.DoesNotExist:
                        dpcmodels.MonthlyBalance.objects.create(
                            date=next_month,
                            source=source,
                            start=0,
                        )
                    next_month = (next_month.replace(day=1) + datetime.timedelta(days=32)).replace(day=1)
                # After the while, create the current month MB, which we know it was missing from
                # the very beginning of this code branch
                # This is to avoid creating next month monthly balances when the source has all the
                # monthly balances (use case: 2 sources, 1 complete and 1 lacking a monthly balance)
                if next_month <= day.replace(day=1):
                    dpcmodels.MonthlyBalance.objects.create(
                        date=next_month,
                        source=source,
                        start=0,
                    )
                # Update all the monthly balances.
                # This can be a bottleneck obviosly, but we don't expect to go down this route very
                # often (if the user does not delete random MB objects will be enough to avoid this
                # problem)
                mbs = dpcmodels.MonthlyBalance.objects\
                    .filter(date__lte=day.replace(day=1), source=source)\
                    .order_by("date")\
                    .select_related()
                for mb in mbs:
                    mb.update_balance()
            else:
                print("ALTRO PROBLEMA!")
    balances_qs = dpcmodels.MonthlyBalance.objects.filter(
        date__month=day.month,
        date__year=day.year,
    ).order_by('source__name')
    # Prepare data about balances and global totals, mainly
    for b in balances_qs:
        bpk = b.pk
        spk = b.source.pk
        # refs #5087 We're adding keys to the OrderedDict, based on balances_qs sort (source__name)
        data['per_source'][spk] = {
            'bpk': bpk,
            'spk': spk,
            'name': b.source.name,
            'start': b.start,
            # TODO: Think about a better fix for the bug when there is a balance with end=None
            'end': b.end if b.end else Decimal('0.00'),
            'inbound': Decimal('0.00'),
            'outbound': Decimal('0.00'),
            'total': Decimal('0.00'),
        }
        data['global']['start'] += data['per_source'][spk]['start']
        data['global']['end'] += data['per_source'][spk]['end']
    # Prepare the data about entries
    data['entries'] = []
    for e in entries:
        this_entry = {
            'entry': e,
            'amount_cells': [],
        }
        for source in sources_qs:
            # This is not the entry's source
            if e.source != source:
                this_entry['amount_cells'].extend(['', ''])
            # This is the entry's source
            else:
                # The entry is inbound
                if e.kind == dpcconsts.ENTRY_KIND_INBOUND:
                    # Extend amount_cells with the entry's data
                    this_entry['amount_cells'].extend([e.amount, ''])
                    # Update the totals
                    data['per_source'][e.source.pk]['inbound'] += e.amount
                    data['per_source'][e.source.pk]['total'] += e.amount
                    data['global']['inbound'] += e.amount
                    data['global']['total'] += e.amount
                # The entry is outbound
                elif e.kind == dpcconsts.ENTRY_KIND_OUTBOUND:
                    # Extend amount_cells with the entry's data
                    this_entry['amount_cells'].extend(['', e.amount])
                    # Update the totals
                    data['per_source'][e.source.pk]['outbound'] += e.amount
                    data['per_source'][e.source.pk]['total'] -= e.amount
                    data['global']['outbound'] += e.amount
                    data['global']['total'] -= e.amount
        data['entries'].append(this_entry)
    # This is sorted on source name
    data['sources'] = list(sources_qs)
    # This is sorted on source name
    data['balances'] = list(balances_qs)
    return data


# CSV reader/writer UTF8 aware: https://docs.python.org/2/library/csv.html
class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def __next__(self):
        return self.reader.next().encode("utf-8")
