# coding: utf-8
# Copyright (c) 2015, 2018 Truelite Srl <info@truelite.it>
# Author: Mark Caglienzi <mark@truelite.it>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
import decimal
import json
import csv
from django import http
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import permission_required
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse_lazy
from django.db.models import Max, Min
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext_lazy as _
from django.views.generic.base import View
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, DeleteView, FormMixin, UpdateView
from django.views.generic.list import ListView
import django_pettycash.consts as dpcconsts
import django_pettycash.forms as dpcforms
import django_pettycash.models as dpcmodels
import django_pettycash.utils as dpcutils


# refs #4782
COMPANY_MODEL = getattr(settings, 'COMPANY_MODEL')
COMPANY_CREATE_VIEW_URL_NAME = getattr(settings, 'COMPANY_CREATE_VIEW_URL_NAME')


# refs #5072
class DPCDateFormatMixin(object):
    def get_context_data(self, **kwargs):
        context = super(DPCDateFormatMixin, self).get_context_data(**kwargs)
        context['JSDATEFORMAT'] = getattr(settings, 'JSDATEFORMAT', 'YYYY-MM-DD')
        return context


# Entry CRUD views
class DPCEntryListView(ListView):
    model = dpcmodels.Entry

    def dispatch(self, *args, **kwargs):
        user_perms = self.request.user.get_all_permissions()
        for perm in ['django_pettycash.add_entry', 'django_pettycash.change_entry', 'django_pettycash.delete_entry']:
            if perm in user_perms:
                return super(DPCEntryListView, self).dispatch(*args, **kwargs)
        raise PermissionDenied()

    def get_queryset(self):
        qs = super(DPCEntryListView, self).get_queryset().select_related()
        # Filter this month's entries
        month = self.kwargs.get('month', '')
        if not month:
            month = datetime.date.today().strftime("%Y-%m")
        filter_range = dpcutils.monthrange(
            dpcutils.month_string_to_date(month)
        )
        qs = qs.filter(date__range=filter_range).order_by('date')
        return qs

    def get_context_data(self, **kwargs):
        month = self.kwargs.get('month', '')
        if not month:
            month = datetime.date.today().strftime("%Y-%m")
        this_month = dpcutils.month_string_to_date(month)
        context = super(DPCEntryListView, self).get_context_data(**kwargs)
        context['month'] = month
        context['data'] = dpcutils.get_entry_list_table_data(
            day=this_month,
            entries=self.object_list,
            request=self.request,
        )
        context['previous_month'] = dpcutils.get_previous_month(this_month)
        next_month = dpcutils.get_next_month(this_month)
        if not dpcutils.is_month_in_the_future(next_month):
            context['next_month'] = dpcutils.get_next_month(this_month)
        else:
            context['next_month'] = None
        context['company_create_url'] = COMPANY_CREATE_VIEW_URL_NAME
        return context


# refs #4794
class DPCEntryCSVListView(DPCEntryListView):
    def render_to_response(self, context, *args, **kwargs):
        csv_filename = "entries-%s.csv" % context['month']
        response = http.HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="%s"' % csv_filename
        writer = csv.writer(response)
        # Month
        writer.writerow([context['data']['month']])
        # Sources
        sources = []
        for source in context['data']['sources']:
            sources.extend([source.name, ''])
        writer.writerow(['', ''] + sources)
        # Initial balances
        initial_balance = [_("Initial balance"), "%0.2f" % context['data']['global']['start']]
        for source_pk, source_data in context['data']['per_source'].items():
            initial_balance.extend(["%0.2f" % source_data['start'], ''])
        writer.writerow(initial_balance)
        writer.writerow([_("Date"), _("Reason")] + [_("Inbound"), _("Outbound")] * len(context['data']['sources']) + [_("Tags")])
        # Entries
        for e in context['data']['entries']:
            row = ["%s" % e['entry'].date, e['entry'].reason]
            row.extend(e['amount_cells'])
            row.append(",".join([t.name for t in e['entry'].tags.all()]))
            writer.writerow(row)
        # Totals
        totals = [_("Totals"), '']
        for source_pk, source_data in context['data']['per_source'].items():
            totals.extend(["%0.2f" % source_data['inbound'], "%0.2f" % source_data['outbound']])
        writer.writerow(totals)
        # refs #4903
        source_totals = ['', '',]
        for source_pk, source_data in context['data']['per_source'].items():
            source_totals.extend(["%0.2f" % source_data['total'], ''])
        writer.writerow(source_totals)
        # Final balances
        final_balances = [_("Final balance - %s" % context['data']['month']), "%0.2f" % context['data']['global']['end']]
        for source_pk, source_data in context['data']['per_source'].items():
            final_balances.extend(["%0.2f" % source_data['end'], ''])
        writer.writerow(final_balances)
        return response


class DPCEntryCreateView(DPCDateFormatMixin, CreateView):
    model = dpcmodels.Entry
    form_class = dpcforms.EntryModelForm

    # refs #5584
    def get_success_url(self):
        month = self.object.date.strftime("%Y-%m")
        return reverse_lazy('dpc_entry_list', kwargs={'month': month})

    # refs #4775
    def form_valid(self, form):
        form.create_monthly_balance_objects_if_needed()
        form.instance.created_by = self.request.user
        form.instance.modified_by = self.request.user
        messages.info(self.request, _("Entry created"))
        return super(DPCEntryCreateView, self).form_valid(form)

    @method_decorator(permission_required('django_pettycash.add_entry', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCEntryCreateView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(DPCEntryCreateView, self).get_context_data(**kwargs)
        context['new_company_form'] = dpcforms.CompanyModelForm()
        context['company_create_url'] = COMPANY_CREATE_VIEW_URL_NAME
        return context


class DPCEntryUpdateView(DPCDateFormatMixin, UpdateView):
    model = dpcmodels.Entry
    form_class = dpcforms.EntryModelForm

    # refs #5584
    def get_success_url(self):
        month = self.object.date.strftime("%Y-%m")
        return reverse_lazy('dpc_entry_list', kwargs={'month': month})

    # refs #4775
    def form_valid(self, form):
        form.create_monthly_balance_objects_if_needed()
        form.instance.modified_by = self.request.user
        messages.info(self.request, _("Entry updated"))
        return super(DPCEntryUpdateView, self).form_valid(form)

    def get_context_data(self, **kwargs):
        context = super(DPCEntryUpdateView, self).get_context_data(**kwargs)
        context['company_create_url'] = COMPANY_CREATE_VIEW_URL_NAME
        return context

    @method_decorator(permission_required('django_pettycash.change_entry', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCEntryUpdateView, self).dispatch(*args, **kwargs)


class DPCEntryDeleteView(DeleteView):
    model = dpcmodels.Entry
    success_url = reverse_lazy('dpc_entry_list')

    @method_decorator(permission_required('django_pettycash.delete_entry', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCEntryDeleteView, self).dispatch(*args, **kwargs)


class DPCEntryDetailView(DetailView):
    model = dpcmodels.Entry

    def dispatch(self, *args, **kwargs):
        user_perms = self.request.user.get_all_permissions()
        for perm in ['django_pettycash.add_entry', 'django_pettycash.change_entry', 'django_pettycash.delete_entry']:
            if perm in user_perms:
                return super(DPCEntryDetailView, self).dispatch(*args, **kwargs)
        raise PermissionDenied()


# refs #4782
class DPCEntryCompanyListAjaxView(View):
    """
    View used by the Entry.company select2 plugin, via AJAX GET call.
    """
    def get(self, request, *args, **kwargs):
        q = request.GET.get('term', '')
        company_model = dpcmodels.get_company_model()
        companies = []
        for c in company_model.objects.all():
            if not q or (q and q.lower() in str(c).lower()):
                companies.append({'id': c.pk, 'text': str(c)})
        results = {'results': companies}
        response = http.HttpResponse(json.dumps(results), content_type='application/json')
        return response


# Source CRUD views
class DPCSourceListView(ListView):
    model = dpcmodels.Source

    def dispatch(self, *args, **kwargs):
        user_perms = self.request.user.get_all_permissions()
        for perm in ['django_pettycash.add_source', 'django_pettycash.change_source', 'django_pettycash.delete_source']:
            if perm in user_perms:
                return super(DPCSourceListView, self).dispatch(*args, **kwargs)
        raise PermissionDenied()


class DPCSourceCreateView(CreateView):
    model = dpcmodels.Source
    form_class = dpcforms.SourceCreateModelForm

    @method_decorator(permission_required('django_pettycash.add_source', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCSourceCreateView, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        # refs #11813
        # Save start_balance here, to allow creation of the Monthly Balance in get_success_url(),
        # where self.object is available
        self.start_balance = form.cleaned_data.get('start_balance')
        return super().form_valid(form)

    def get_success_url(self):
        # refs #11813
        dpcmodels.MonthlyBalance.objects.create(
            source=self.object,
            date=datetime.date.today().replace(day=1),
            start=self.start_balance,
            end=self.start_balance,
        )
        messages.info(self.request, _("Source and starting Monthly Balance created"))
        return reverse_lazy('dpc_source_list')


class DPCSourceUpdateView(UpdateView):
    model = dpcmodels.Source
    form_class = dpcforms.SourceModelForm
    success_url = reverse_lazy('dpc_source_list')

    @method_decorator(permission_required('django_pettycash.change_source', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCSourceUpdateView, self).dispatch(*args, **kwargs)


class DPCSourceDeleteView(DeleteView):
    model = dpcmodels.Source
    success_url = reverse_lazy('dpc_source_list')

    @method_decorator(permission_required('django_pettycash.delete_source', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCSourceDeleteView, self).dispatch(*args, **kwargs)


class DPCSourceDetailView(DetailView):
    model = dpcmodels.Source

    def dispatch(self, *args, **kwargs):
        user_perms = self.request.user.get_all_permissions()
        for perm in ['django_pettycash.add_source', 'django_pettycash.change_source', 'django_pettycash.delete_source']:
            if perm in user_perms:
                return super(DPCSourceDetailView, self).dispatch(*args, **kwargs)
        raise PermissionDenied()


# Tag CRUD views
class DPCTagListView(ListView):
    model = dpcmodels.Tag

    def dispatch(self, *args, **kwargs):
        user_perms = self.request.user.get_all_permissions()
        for perm in ['django_pettycash.add_tag', 'django_pettycash.change_tag', 'django_pettycash.delete_tag']:
            if perm in user_perms:
                return super(DPCTagListView, self).dispatch(*args, **kwargs)
        raise PermissionDenied()


class DPCTagCreateView(CreateView):
    model = dpcmodels.Tag
    form_class = dpcforms.TagModelForm
    success_url = reverse_lazy('dpc_tag_list')

    @method_decorator(permission_required('django_pettycash.add_tag', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCTagCreateView, self).dispatch(*args, **kwargs)


class DPCTagUpdateView(UpdateView):
    model = dpcmodels.Tag
    form_class = dpcforms.TagModelForm
    success_url = reverse_lazy('dpc_tag_list')

    @method_decorator(permission_required('django_pettycash.change_tag', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCTagUpdateView, self).dispatch(*args, **kwargs)


class DPCTagDeleteView(DeleteView):
    model = dpcmodels.Tag
    success_url = reverse_lazy('dpc_tag_list')

    @method_decorator(permission_required('django_pettycash.delete_tag', raise_exception=True))
    def dispatch(self, *args, **kwargs):
        return super(DPCTagDeleteView, self).dispatch(*args, **kwargs)


class DPCTagDetailView(DetailView):
    model = dpcmodels.Tag


# refs #4921
class DPCStatsBalanceByTag(DPCDateFormatMixin, ListView, FormMixin):
    form_class = dpcforms.StatsBalanceByTagForm
    model = dpcmodels.Entry
    template_name = 'django_pettycash/statistics_balance_by_tag.html'
    success_url = reverse_lazy('home')

    def get_queryset(self):
        # If we came from a clean form (POST), do the filtering
        if hasattr(self.form, 'cleaned_data'):
            qs = self.model.objects.all().select_related()
            tags = self.form.cleaned_data.get('tags', dpcmodels.Tag.objects.none())
            from_date = self.form.cleaned_data.get('from_date', None)
            to_date = self.form.cleaned_data.get('to_date', None)
            # Avoid getting duplicate entries, which is sometimes possible when
            # m2m and __in filtering are involved, using distinct()
            qs = qs.filter(tags__in=tags).distinct()
            if from_date:
                qs = qs.filter(date__gte=from_date)
            if to_date:
                qs = qs.filter(date__lte=to_date)
        # Else don't show anything
        else:
            qs = self.model.objects.none()
        return qs

    def get_context_data(self, **kwargs):
        context = super(DPCStatsBalanceByTag, self).get_context_data(**kwargs)
        # Calculate the sums here, after possible filtering
        data = {}
        context['inbound_total'] = decimal.Decimal('0.00')
        context['outbound_total'] = decimal.Decimal('0.00')
        context['inbound_count'] = 0
        context['outbound_count'] = 0
        for source in dpcmodels.Source.objects.all():
            inbound_entries = self.object_list.filter(
                source=source,
                kind=dpcconsts.ENTRY_KIND_INBOUND,
            )
            outbound_entries = self.object_list.filter(
                source=source,
                kind=dpcconsts.ENTRY_KIND_OUTBOUND,
            )
            data[source.name] = {
                'source_name': source.name,
                'inbound_count': inbound_entries.count(),
                'inbound_total': sum([x['amount'] for x in inbound_entries.values('amount')]) or decimal.Decimal('0.00'),
                'outbound_count': outbound_entries.count(),
                'outbound_total': sum([x['amount'] for x in outbound_entries.values('amount')]) or decimal.Decimal('0.00'),
            }
            context['inbound_total'] += data[source.name]['inbound_total']
            context['outbound_total'] += data[source.name]['outbound_total']
            context['inbound_count'] += data[source.name]['inbound_count']
            context['outbound_count'] += data[source.name]['outbound_count']
            data[source.name]['total'] = data[source.name]['inbound_total'] - data[source.name]['outbound_total']
            data[source.name]['count'] = data[source.name]['inbound_count'] + data[source.name]['outbound_count']
        context['data'] = data
        context['min_date'] = self.object_list.aggregate(Min('date'))['date__min'] or ''
        context['max_date'] = self.object_list.aggregate(Max('date'))['date__max'] or ''
        context['total'] = context['inbound_total'] - context['outbound_total']
        context['count'] = context['inbound_count'] + context['outbound_count']
        return context

    def get(self, request, *args, **kwargs):
        self.form = self.get_form(self.form_class)
        self.object_list = self.get_queryset()
        context = self.get_context_data(object_list=self.object_list, form=self.form)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        self.form = self.form_class(request.POST)
        # Trigger form validation and create self.form.cleaned_data dict by calling is_valid()
        if self.form.is_valid():
            self.object_list = self.get_queryset()
        context = self.get_context_data(object_list=self.object_list, form=self.form)
        return self.render_to_response(context)
